/*
FILE:     fileobj.h
*/
/*
VERSION:  2.053
*/
/*
DATE:     1/24/2007
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2007 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#ifndef _FILEOBJ_H_
#define _FILEOBJ_H_

#include <stdio.h>

#define EXPERIMENT_TYPE_BASIC       1
#define EXPERIMENT_TYPE_XRAY        2
#define EXPERIMENT_TYPE_NMR         4
#define EXPERIMENT_TYPE_MODEL       8
#define EXPERIMENT_TYPE_NEUTRON    16
#define EXPERIMENT_TYPE_ELECTRON   32

#define NDB_DATE_FIELD            7

#include "token_card.h"
#include "molecule.h"
#include "CifFileObj.h"
#include "cry_util.h"
#include "remark.h"

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

typedef struct {
        char FstAtom[5];
        char altLoc1[3];
        int  FstIndex;
        int  FstRes;
        char SndAtom[5];
        char altLoc2[3];
        int  SndIndex;
        int  SndRes;
        char SymOP_1[7];
        char SymOP_2[7];
        char *details;
} _LINK;

typedef struct {
        char sheetID[4];
        int  numStrands;
        int  initIndex;
        int  initRes;
        int  endIndex;
        int  endRes;
        int  sense;
        char curAtom[5];
        int  curIndex;
        int  curRes;
        char prevAtom[5];
        int  prevIndex;
        int  prevRes;
} _SHEET;

typedef struct {
        int  FstIndex;
        int  FstRes;
        int  SndIndex;
        int  SndRes;
        char SymOP_1[7];
        char SymOP_2[7];
} _SSBOND;

typedef struct {
        char FstAtom[5];
        char altLoc1[3];
        int  FstIndex;
        int  FstRes;
        char H_Atom[5];
        char altLocH[3];
        int  H_Index;
        int  H_Res;
        char SndAtom[5];
        char altLoc2[3];
        int  SndIndex;
        int  SndRes;
        char SymOP_1[7];
        char SymOP_2[7];
} _HYDBND;

typedef _LINK _SLTBRG;

typedef struct {
        char SiteID[4];
        int  numRes;
        int  alloc_numRes;
        int  *initIndex;
        int  *initRes;
} _SITE;

typedef struct {
        char PDB_ID[5];
        int  FstIndex;
        int  FstRes;
        char Standard_Name[4];
        char *details;
} _MODRES;

typedef struct {
        int  initIndex;
        int  initRes;
        int  numAtoms;
} _HET;

class _FileObj
{
   public:
      _FileObj(const int, const int);
      ~_FileObj();
      void  set_input_filename(const char*);
      void  set_output_filename(const char*);
      int   read_mmcif_file();
      void  cif_to_pdb(int &chain_flag, const int iupac_flag);
      void  write_pdb_file(const int chain_flag, const int iupac_flag);

   private:
      void _ManageSpace();
      void _ManageSeqSpace();
      void _ManageResSpace(SEQ *seq);

      void put_coordinates_into_molecules();
      void cifndb_process_asym_poly_seq_ndb();
      void match_entity_poly_seq_to_seqres();
      void insert_to_seqs(ISTable *Table, char *entity_id, const int SeqLen,
              char **seqs);

      void print_general_cards(FILE *fp, const int card_id, int &numRemark);
      void print_jrnl_cards(FILE *fp, int &numRemark);
      void print_molecules(FILE *fp, const int chain_flag, const int iupac_flag);
      void print_atom(FILE *fp, const int card_id, int &atomSerialNo,
             int &WaterSerialNo, _ATOM *atom, char FieldInfo[15][10],
             const int n_mol, const int chain_flag);
      void print_master_card(FILE *fp, const int numRemark);
      int  ndb_count_general_line_no(const int card_id);
      int  ndb_count_atom_line_no();
      int  ndb_count_num_chain();

      void get_cell();
      void delete_null_cards();
      void update_tokens();
      void get_sequence_types();
      void get_seqres_from_seqs(int &chain_flag);
      void Insert_a_Sheet(const char *id, const int numStrands, const int initIndex,
               const int initRes, const int endIndex, const int endRes, const int sense,
               const char *curAtom, const int curIndex, const int curRes,
               const char *prevAtom, const int prevIndex, const int prevRes);
      void Insert_a_SSbond(const int FstIndex, const int FstRes, const int SndIndex,
                  const int SndRes, const char *SymOP_1, const char *SymOP_2);
      void Insert_a_Hydbnd(const char *FstAtom, const char *altLoc1, const int FstIndex,
           const int FstRes, const char *H_Atom, const char *altLocH, const int H_Index,
           const int H_Res, const char *SndAtom, const char *altLoc2, const int SndIndex,
           const int SndRes, const char *SymOP_1, const char *SymOP_2); 
      void Insert_a_Link(const char *FstAtom, const char *altLoc1, const int FstIndex,
           const int FstRes, const char *SndAtom, const char *altLoc2, const int SndIndex,
           const int SndRes, const char *SymOP_1, const char *SymOP_2, const char *details);
      void Insert_a_Sltbrg(const char *FstAtom, const char *altLoc1, const int FstIndex,
           const int FstRes, const char *SndAtom, const char *altLoc2, const int SndIndex,
           const int SndRes, const char *SymOP_1, const char *SymOP_2);
      void Insert_a_Site(const char *SiteID, const int initIndex, const int initRes);
      void Insert_a_Het(const int initIndex, const int initRes, const int numAtoms);
      void Insert_a_Modres(const char *PDB_ID, const int FstIndex, const int FstRes,
            const char *Standard_Name, const char *details);
      void output_structural_features(const int chain_flag);

      void processing_r_factor();
      void cifndb_process_category(ISTable *Table, const int id, const int is_special);
      void cifndb_process_general(const int tokenid, const int fieldno,
              const int NumKey, int *SerialNoField, char SerialNo[10][82],
              char *value, const int IsJrnl);
      void cifndb_process_special(const int tokenid, const int fieldno,
              const int NumKey, int *SerialNoField, char SerialNo[10][82],
              char *value);
      void cifndb_process_general_category(ISTable *Table, const int id);
      void cifndb_process_refine_ls_restr(ISTable *Table);
      void cifndb_process_matrix(ISTable *Table, const int id);
      void cifndb_process_struct_site_gen(ISTable *Table);
      void cifndb_process_rcsb_link(ISTable *Table);
      void cifndb_process_struct_conf(ISTable *Table);
      void cifndb_process_sheet();
      void cifndb_process_database_pdb_rev(ISTable *Table);
      void cifndb_process_struct_conn(ISTable *Table);
      void put_into_card(char *value, const int tokenid, const int fieldno);
      void post_processing_from_cif_to_ndb();
      void get_het_card();
      void insert_formul_card(const char *hetID, int &serialNo, const int count, const int num);
      void insert_het_card(_CHAIN *chain, const char *resname,
               const int resnum, const char *ins_code, const int atmnum);

      void ndb_to_pdb_processing_authors();
      void ndb_to_pdb_processing_shelx_and_tnt();
      void ndb_to_pdb_processing_synchrotron();
      void ndb_to_pdb_proc_header();
      void replace_pdb_compnd_source_card();
      void add_compnd_card(const char *value, const int serialNo);
      void ndb_to_pdb_get_source_card(const int serialno, const int entityId,
              const int isSynthetic, int *sourceSerialNo, const int nsource,
              PDB_Source_Format *PdbSource);
      void ndb_to_pdb_get_remarks(const int iupac_flag);
      int  ndb_to_pdb_get_remark_3_front();
      void ndb_to_pdb_get_group_remarks(const int program, const int remark_no);
      void ndb_to_pdb_get_general_non_loop_remark(const int num_remarks, REMARKS *Remarks);
      int  ndb_to_pdb_get_remark_from_premark(const int remarkNo);
      void ndb_to_pdb_get_remark_2();
      void ndb_to_pdb_get_remark_3_loop_xplor(const int num_remarks, REMARKS *Remarks);
      void ndb_to_pdb_get_remark_3_loop_xplor1(const int num_remarks, REMARKS *Remarks);
      void ndb_to_pdb_get_remark_3_loop_xplor2(const int num_remarks, REMARKS *Remarks);
      void ndb_to_pdb_get_remark_3_loop_xplor_details(REMARKS *Remarks,
             const int remark_no, One_Card_Info_Format *pCard,
             const int index, const int card_id);
      void ndb_to_pdb_get_remark_280_crystal();
      void ndb_to_pdb_get_remark_290();
      int  ndb_to_pdb_get_remark_300();
      int  ndb_to_pdb_get_remark_350();
      void ndb_to_pdb_processing_nmr_remarks();
      int  check_virus_ncs_remark_300();
      int  check_virus_ncs_remark_350();
      void close_contacts();
      void rms_deviation_of_bonds();
      void rms_deviation_of_angles();
      void outside_torsions();
      void non_cis_trans_torsions();
      void output_close_contact(ISTable*, BTreeObj<_DIndex>&, const int);
      void construct_remark_900();
      void check_atom_number_count();
      int  get_matrix(double m[3][4], const int serial_no);

      char               _input_filename[200];
      char               _output_filename[200];
      char               _temp_file[200];
      char               _StructureId[100];
      int                _experiment_type;
      int                _num_mol;
      int                _alloc_mol;
      Cards_Info_Format* _CardInfo;
      Cards_Info_Format* _JrnlCardInfo;
      _MOLECULE**        _molecules;
      CifFileObj*        _CifObj;
      int                _Alloc_Seqs;
      int                _Num_Seqs;
      SEQ**              _Seqs;
      NDBCELLCONST*      _cell;
      CifString          _error_message;

      int                _alloc_link, _n_link;
      _LINK**            _link;
      int                _alloc_sheet, _n_sheet;
      _SHEET**           _sheet;
      int                _alloc_ssbond, _n_ssbond;
      _SSBOND**          _ssbond;
      int                _alloc_hydbnd, _n_hydbnd;
      _HYDBND**          _hydbnd;
      int                _alloc_sltbrg, _n_sltbrg;
      _SLTBRG**          _sltbrg;
      int                _alloc_site, _n_site;
      _SITE**            _site;
      int                _alloc_het, _n_het;
      _HET**             _het;
      int                _alloc_modres, _n_modres;
      _MODRES**          _modres;
};

#endif
