/*
FILE:     cif_to_ndb.C
*/
/*
VERSION:  2.053
*/
/*
DATE:     1/24/2007
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2007 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <string.h>

#include "ndb_misclib.h"
#include "util.h"
#include "fileobj.h"
#include "erf_iolib.h"
#include "erf_util.h"

#define N_R_FACTOR  16

typedef struct {
        const char *catetory;
        const char *item;
} R_FACTOR;

typedef struct {
        int  num;
        char date[20];
        char replaces[10];
        int  mod_type;
        char *record;
} PDB_REV;

static R_FACTOR r_factor[N_R_FACTOR] = {
        { "ndb_refine",      "R_factor_all_4sig_cutoff" },
        { "ndb_refine",      "R_factor_all_no_cutoff" },
        { "ndb_refine",      "R_factor_obs_4sig_cutoff" },
        { "ndb_refine",      "R_factor_obs_no_cutoff" },
        { "ndb_refine",      "free_R_factor_4sig_cutoff" },
        { "ndb_refine",      "free_R_factor_no_cutoff" },
        { "refine",          "ls_R_factor_R_free" },
        { "refine",          "ls_R_factor_R_work" },
        { "refine",          "ls_R_factor_all" },
        { "refine",          "ls_R_factor_obs" },
        { "refine_ls_shell", "R_factor_R_free" },
        { "refine_ls_shell", "R_factor_R_work" },
        { "reflns",          "ndb_Rmerge_I_obs" },
        { "reflns",          "ndb_Rsym_value" },
        { "reflns_shell",    "Rmerge_I_obs" },
        { "reflns_shell",    "ndb_Rsym_value" }
};

static int get_number_strands(ISTable *Table, const char *sheet_id);
static int get_sense(ISTable *Table, const char *sheet_id, const char *id_1, const char *id_2);
static void get_hbond(_MOLECULE *mol, ISTable *Table, const int chain_index,
            const char *sheet_id, const char *range_id, CifString &curAtom, int &curIndex,
            int &curRes, CifString &prevAtom, int &prevIndex, int &prevRes);
static void cif_to_ndb_find_serial_no(ISTable *Table, const int id, int irow,
         int tokenid, int *idfield, char idvalue[10][82], char SerialNo[10][82]);
#ifdef SUN_OS
#ifdef __cplusplus
extern "C" {
#endif
#endif
static int pdb_rev_compare(const void *a, const void *b);
#ifdef SUN_OS
#ifdef __cplusplus
}
#endif
#endif
static int get_residue_index(_MOLECULE *mol, ISTable *Table, const int irow,
                const int ndb_comp_id, const int ndb_asym_id, const int ndb_seq_id,
                const int pdb_comp_id, const int pdb_asym_id, const int pdb_seq_id,
                const int ins_code_id, int &FstIndex, int &FstRes);

void _FileObj::cif_to_pdb(int& chain_flag, const int iupac_flag)
{
       ISTable  *Table = NULL, *Struct_Ref = NULL;
       int i, j, rowNo, colNo, struct_ref_id = -1;
       int auth_beg_idx, auth_end_idx, beg_idx, end_idx;
       int auth_beg_chn_idx, auth_end_chn_idx, beg_chn_idx, end_chn_idx;
       CifString string;

       processing_r_factor();
       Table = _CifObj->GetTablePtr(_StructureId, "ndb_poly_seq_scheme");
       if (!is_empty_table(Table)) {
            rowNo = Table->GetNumRows();
            colNo = Table->GetNumColumns();
            for (i = 0; i < rowNo; i++) {
                 for (j = 0; j < colNo; j++) {
                      Table->GetCell(string, j, i);
                      if (string == "n/a") {
                           string.Copy("...");
                           Table->UpdateCell(string, j, i);
                      }
                 }
            }
       }
       Table = _CifObj->GetTablePtr(_StructureId, "struct_ref_seq");
       if (!is_empty_table(Table)) {
            rowNo = Table->GetNumRows();
            auth_beg_idx = Table->GetColumnIndex("ndb_auth_seq_align_beg");
            auth_end_idx = Table->GetColumnIndex("ndb_auth_seq_align_end");
            beg_idx = Table->GetColumnIndex("seq_align_beg");
            end_idx = Table->GetColumnIndex("seq_align_end");
            for (i = 0; i < rowNo; i++) {
                 Table->GetCell(string, auth_beg_idx, i);
                 Table->UpdateCell(string, beg_idx, i);
                 Table->GetCell(string, auth_end_idx, i);
                 Table->UpdateCell(string, end_idx, i);
            }
       }
       Table = _CifObj->GetTablePtr(_StructureId, "struct_ref_seq_dif");
       if (!is_empty_table(Table)) {
            rowNo = Table->GetNumRows();
            auth_beg_idx = Table->GetColumnIndex("ndb_auth_seq_num");
            beg_idx = Table->GetColumnIndex("seq_num");
            for (i = 0; i < rowNo; i++) {
                 Table->GetCell(string, auth_beg_idx, i);
                 Table->UpdateCell(string, beg_idx, i);
            }
       }
       Table = _CifObj->GetTablePtr(_StructureId, "struct_mon_prot_cis");
       if (!is_empty_table(Table)) {
            rowNo = Table->GetNumRows();
            auth_beg_idx = Table->GetColumnIndex("auth_seq_id");
            auth_beg_chn_idx = Table->GetColumnIndex("auth_asym_id");
            auth_end_idx = Table->GetColumnIndex("ndb_auth_seq_id_2");
            auth_end_chn_idx = Table->GetColumnIndex("ndb_auth_asym_id_2");
            beg_idx = Table->GetColumnIndex("label_seq_id");
            beg_chn_idx = Table->GetColumnIndex("label_asym_id");
            end_idx = Table->GetColumnIndex("ndb_label_seq_id_2");
            end_chn_idx = Table->GetColumnIndex("ndb_label_asym_id_2");
            for (i = 0; i < rowNo; i++) {
                 Table->GetCell(string, auth_beg_idx, i);
                 Table->UpdateCell(string, beg_idx, i);
                 Table->GetCell(string, auth_beg_chn_idx, i);
                 Table->UpdateCell(string, beg_chn_idx, i);
                 Table->GetCell(string, auth_end_idx, i);
                 Table->UpdateCell(string, end_idx, i);
                 Table->GetCell(string, auth_end_chn_idx, i);
                 Table->UpdateCell(string, end_chn_idx, i);
            }
       }

       for (i = 0; i < MxCategory; i++) {
            if (!strcmp(CifCategory[i].category, "struct_conf_type") ||
                !strcmp(CifCategory[i].category, "struct_conn_type") ||
                !strcmp(CifCategory[i].category, "struct_sheet_order") ||
                !strcmp(CifCategory[i].category, "struct_sheet_range") ||
                !strcmp(CifCategory[i].category, "struct_sheet_hbond") ||
                !strcmp(CifCategory[i].category, "rcsb_struct_sheet_hbond") ||
                !strcmp(CifCategory[i].category, "atom_site") ||
                !strcmp(CifCategory[i].category, "atom_site_anisotrop") ||
                !strcmp(CifCategory[i].category, "struct_asym") ||
                !strcmp(CifCategory[i].category, "entity_poly_seq") ||
                !strcmp(CifCategory[i].category, "ndb_nonstandard_list") ||
                !strcmp(CifCategory[i].category, "ndb_poly_seq_scheme") ||
                !strcmp(CifCategory[i].category, "database_PDB_rev_record") ||
                !strcmp(CifCategory[i].category, "ndb_database_remark") ||
                !strcmp(CifCategory[i].category, "database_PDB_remark") ||
                !strcmp(CifCategory[i].category, "ndb_pdb_compnd") ||
                !strcmp(CifCategory[i].category, "ndb_pdb_source"))
                 continue;

            Table = _CifObj->GetTablePtr(_StructureId, CifCategory[i].category);
            if (is_empty_table(Table)) continue;

            if (!strcmp(CifCategory[i].category, "struct_ref")) {
                 Struct_Ref = Table;
                 struct_ref_id = i;
                 continue;
            } if (!strcmp(CifCategory[i].category, "struct_conf"))
                 cifndb_process_struct_conf(Table);
            else if (!strcmp(CifCategory[i].category, "struct_sheet"))
                 cifndb_process_sheet();
            else if (!strcmp(CifCategory[i].category, "struct_conn"))
                 cifndb_process_struct_conn(Table);
            else if (!strcmp(CifCategory[i].category, "refine_ls_restr")) 
                 cifndb_process_refine_ls_restr(Table);
            else if (!strcmp(CifCategory[i].category, "struct_site_gen")) 
                 cifndb_process_struct_site_gen(Table);
            else if (!strcmp(CifCategory[i].category, "rcsb_struct_link"))
                 cifndb_process_rcsb_link(Table);
            else if (!strcmp(CifCategory[i].category, "database_PDB_rev"))
                 cifndb_process_database_pdb_rev(Table);
            else if (CifCategory[i].general)
                 cifndb_process_general_category(Table, i);
            else if (CifCategory[i].xray_or_nmr & 4)
                 cifndb_process_matrix(Table, i);
            else cifndb_process_category(Table, i, 0);
       }
       if (Struct_Ref) cifndb_process_category(Struct_Ref, struct_ref_id, 1);

       post_processing_from_cif_to_ndb();

       get_seqres_from_seqs(chain_flag);

       get_het_card();
       update_tokens();
       output_structural_features(chain_flag);
       
       One_Card_Info_Format *pCard = _CardInfo[CRYST1_TOKEN - 1].Head();
       if (pCard) {
            if (strcmp(pCard->Field(10), ""))
                 pCard->AddElementToField(pCard->Field(10), 7);
            if (!strcmp(pCard->Field(7), "C 2"))
                 pCard->AddElementToField("C 1 2 1", 7);
            if (!strcmp(pCard->Field(7), "P 2"))
                 pCard->AddElementToField("P 1 2 1", 7);
            if (!strcmp(pCard->Field(7), "P 21"))
                 pCard->AddElementToField("P 1 21 1", 7);
       }

       ndb_to_pdb_processing_authors();
       ndb_to_pdb_processing_shelx_and_tnt();
       ndb_to_pdb_processing_synchrotron();
       ndb_to_pdb_proc_header();
       replace_pdb_compnd_source_card();
       ndb_to_pdb_get_remarks(iupac_flag);
}

void _FileObj::processing_r_factor()
{
       ISTable *Table = NULL;
       double temp = 0;
       int i = 0, rowNo = 0, item_index = 0, irow = 0;
       CifString string;
       char value[80];

       for (i = 0; i < N_R_FACTOR; i++) {
            if (!_CifObj->IsTablePresent(_StructureId, r_factor[i].catetory))
                 continue;
            Table = _CifObj->GetTablePtr(_StructureId, r_factor[i].catetory);
            rowNo = Table->GetNumRows();
            item_index = Table->GetColumnIndex(r_factor[i].item);

            for (irow = 0; irow < rowNo; irow++) {
                 if (!Table->GetCell(string, item_index, irow)) {
                      if (string != "" && string != "?" && string != ".") {
                           temp = atof(string.Text()) * 100.0;
                           sprintf(value, "%.7f", temp);
                           string.Copy(value);
                           Table->UpdateCell(string, item_index, irow);
                      }
                 }
            }
            _CifObj->WriteTable(Table, _StructureId, r_factor[i].catetory);
       }
}

void _FileObj::cifndb_process_category(ISTable *Table, const int id, const int is_special)
{
       int colNo = Table->GetNumColumns();
       int rowNo = Table->GetNumRows();

       ReVarCifArray<CifString> *ColumnNames = Table->GetColumnNames();

       int  i, j, k, irow, icol, idfield[10], SerialNoField[10], cols[200];
       int tokenid, jrnlid, fieldno;
       char SerialNo[10][82], idValue[10][82];
       CifString string;

       for (i = 0; i < 10; i++) idfield[i] = -1;
       for (i = 0; i < colNo; i++) {
            for (j = 0; j < CifCategory[id].nkeywords; j++) {
                 if (!strcasecmp((*ColumnNames)[i].Text(),
                        CifCategory[id].keywords[j].name)) break;
            }
            if (j == CifCategory[id].nkeywords) cols[i] = -1;
            else                                cols[i] = j;
            for (k = 0; k < CifCategory[id].nCifKeyField; k++) {
                 if (j == CifCategory[id].CifKeyField[k]) idfield[k] = i;
            }
       }
       if (ColumnNames) delete ColumnNames;

       for (irow = 0; irow < rowNo; irow++) {
            for (icol = 0; icol < colNo; icol++) {
                 j = cols[icol];
                 if (j < 0 || j > CifCategory[id].nkeywords) continue;
                 Table->GetCell(string, icol, irow);
                 if (string == "" || string == "." || string == "?")
                      continue;

                 for (i = 0; i < CifCategory[id].keywords[j].numNdbInfo; i++) {
                      tokenid=CifCategory[id].keywords[j].NdbInfo[i].NdbTokenId-1;
                      jrnlid=CifCategory[id].keywords[j].NdbInfo[i].JrnlTokenId-1;
                      fieldno=CifCategory[id].keywords[j].NdbInfo[i].NdbFieldNo-1;
                      for (k = 0; k < CifCategory[id].nCifKeyField; k++)
                           SerialNoField[k] = CifCategory[id].keywords[j].
                                       NdbInfo[i].NdbTokenKeyFieldNo[k];

                      if (tokenid < 0 || tokenid >= MxTokens || fieldno <= 0) continue;
                      cif_to_ndb_find_serial_no(Table, id, irow, tokenid, idfield,
                                       idValue, SerialNo);

                      if (CifCategory[id].nCifKeyField) {
                           for (k = 0; k < CifCategory[id].nCifKeyField; k++) {
                                if (strcmp(idValue[k], "") && strcasecmp(idValue[k],
                                    CifCategory[id].keywords[j].NdbInfo[i].
                                       JrnlTokenName_Type)) break;
                           }
                           if (k < CifCategory[id].nCifKeyField) continue;
                      }
        
                      if (tokenid == JRNL_TOKEN - 1) {
                           if (!strcasecmp(JrnlToken[jrnlid].TokenName, "AUTH") &&
                               !strcasecmp(CifCategory[id].category, "citation"))
                                continue;

                           cifndb_process_general(jrnlid, fieldno, 
                              CifCategory[id].nCifKeyField, SerialNoField, 
                                 SerialNo, string.Text(), TRUE);
                      } else if (is_special)
                           cifndb_process_special(tokenid, fieldno,
                                   CifCategory[id].nCifKeyField,
                                   SerialNoField, SerialNo, string.Text());
                      else cifndb_process_general(tokenid, fieldno, 
                                   CifCategory[id].nCifKeyField, 
                                   SerialNoField, SerialNo, string.Text(), FALSE);
                 }
            }
       }
}

void _FileObj::cifndb_process_general(const int tokenid, const int fieldno, 
	      const int NumKey, int *SerialNoField, char SerialNo[10][82], 
              char *value, const int IsJrnl)
{
       One_Card_Info_Format *pCard = NULL;
       int vField = 5, i, jrnl_id;
       char buffer[1000];

       if (tokenid == NDBFIL_TOKEN - 1 &&
          !strcasecmp(value, "DEPOSIT-0")) {
            pCard = _CardInfo[tokenid].Head();
            if (pCard && strcmp(pCard->Field(fieldno), "")) return;
       }

       if (!IsJrnl && !NdbToken[tokenid].FieldList[fieldno].MixedCase)
            ndb_set_string_to_upper(value);

       if (IsJrnl) {
            pCard = _JrnlCardInfo[tokenid].Head();
            while (pCard) {
                 if (NumKey) {
                      for (i = 0; i < NumKey; i++) {
                           if (SerialNoField[i] > 0 && strcasecmp(SerialNo[i],
                               pCard->Field(SerialNoField[i]))) break;
                      }
                      if (i == NumKey) break;
                 } else if (JrnlToken[tokenid].SeqField != 0 && !strcasecmp(SerialNo[0],
                   pCard->Field(JrnlToken[tokenid].SeqField-1)))
                      break;
                 else if (!strcmp(pCard->Field(fieldno), "")) break;
                 pCard = pCard->Next();
            }
       } else if (tokenid == AUTHOR_TOKEN-1)
            pCard = _CardInfo[tokenid].Tail();
       else {
            pCard = _CardInfo[tokenid].Head();
            while (pCard) {
                 if (NumKey) {
                      for (i = 0; i < NumKey; i++) {
                           if (SerialNoField[i] > 0 && strcasecmp(SerialNo[i], 
                               pCard->Field(SerialNoField[i]))) break;
                      }
                      if (i == NumKey) break;
                 } else if (NdbToken[tokenid].SeqField != 0 && !strcasecmp(SerialNo[0],
                   pCard->Field(NdbToken[tokenid].SeqField-1)))
                      break;
                 else if ((tokenid == HEADER_TOKEN - 1 ||
                           tokenid == PDBFIL_TOKEN - 1 ||
                           tokenid == NDBFIL_TOKEN - 1) &&
                          !strcasecmp(value, pCard->Field(fieldno))) break;
                 else if (!strcmp(pCard->Field(fieldno), "")) break;
                 pCard = pCard->Next();
            }
       }

       if (IsJrnl && pCard == NULL) {
            pCard = _JrnlCardInfo[tokenid].add_a_new_card(JrnlToken[tokenid], 1);
            if (NumKey) {
                 for (i = 0; i < NumKey; i++) {
                      if (SerialNoField[i] > 0)
                           pCard->AddElementToField(SerialNo[i], SerialNoField[i]);
                 }
            } else if (atoi(SerialNo[0]) != 0 && JrnlToken[tokenid].SeqField != 0)
                 pCard->AddElementToField(SerialNo[0], JrnlToken[tokenid].SeqField-1);
       } else if (!IsJrnl && pCard == NULL) {
            pCard = _CardInfo[tokenid].add_a_new_card(NdbToken[tokenid], 0);
            if (NumKey) {
                 for (i = 0; i < NumKey; i++) {
                      if (SerialNoField[i] > 0) 
                           pCard->AddElementToField(SerialNo[i], SerialNoField[i]);
                 }
            } else if (strcmp(SerialNo[0], "") && NdbToken[tokenid].SeqField != 0)
                 pCard->AddElementToField(SerialNo[0], NdbToken[tokenid].SeqField-1);
       }

       int logical = FALSE;
       for (i = 0; i < NumKey; i++) {
            if (fieldno == SerialNoField[i]) logical = TRUE;
       }
       if (!IsJrnl && (logical || fieldno == NdbToken[tokenid].SeqField - 1) ||
           IsJrnl && fieldno == JrnlToken[tokenid].SeqField -1) {}
       else if (!IsJrnl && !strcmp(NdbToken[tokenid].TokenName, "AUTHOR")) {
            ndb_name_conversion_first_last(value);
            if (!strcmp(pCard->Field(fieldno), ""))
                 strcpy(buffer, value);
            else sprintf(buffer, "%s, %s", pCard->Field(fieldno), value);
            pCard->AddElementToField(buffer, fieldno);
       } else if (!IsJrnl || IsJrnl && (!strcmp(pCard->Field(fieldno), "") ||
            strcmp(JrnlToken[tokenid].TokenName, "AUTH") &&
            strcmp(JrnlToken[tokenid].TokenName, "EDIT"))) {
            if (IsJrnl) {
                 jrnl_id = find_jrnl_id("REF");
                 if (jrnl_id == tokenid &&  fieldno == vField + 1)
                      pCard->AddElementToField("V.", vField);
                 if (!strcmp(JrnlToken[tokenid].TokenName, "AUTH") ||
                     !strcmp(JrnlToken[tokenid].TokenName, "EDIT"))
                      ndb_name_conversion_first_last(value);
            }
            pCard->AddElementToField(value, fieldno);
       } else if (atoi(SerialNo[0]) == 
           atoi(pCard->Field(JrnlToken[tokenid].SeqField-1))) {
            ndb_name_conversion_first_last(value);
            sprintf(buffer, "%s, %s", pCard->Field(fieldno), value);
            pCard->AddElementToField(buffer, fieldno);
       }
}

void _FileObj::cifndb_process_special(const int tokenid, const int fieldno,
              const int NumKey, int *SerialNoField, char SerialNo[10][82],
              char *value)
{
       if (!NdbToken[tokenid].FieldList[fieldno].MixedCase)
            ndb_set_string_to_upper(value);
       int i;
       One_Card_Info_Format *pCard = _CardInfo[tokenid].Head();
       while (pCard) {
            for (i = 0; i < NumKey; i++) {
                 if (SerialNoField[i] > 0 && strcmp(SerialNo[i],
                       pCard->Field(SerialNoField[i]))) break;
            }
            if (i == NumKey) pCard->AddElementToField(value, fieldno);
            pCard = pCard->Next();
       }
}

void _FileObj::cifndb_process_general_category(ISTable *Table, const int id)
{
       int colNo = Table->GetNumColumns();
       int rowNo = Table->GetNumRows();

       ReVarCifArray<CifString> *ColumnNames = Table->GetColumnNames();
       int *cols = new int[colNo];

       int ndb_tokenid = -1, ndb_fieldno = -1, tmp_id = -1, i = 0, j = 0;
       int row = 0, col = 0;
       char value[10000];
       CifString string;
       One_Card_Info_Format *pCard = NULL;
       for (i = 0; i < colNo; i++) {
            for (j = 0; j < CifCategory[id].nkeywords; j++) {
                 if (!strcasecmp((*ColumnNames)[i].Text(),
                        CifCategory[id].keywords[j].name)) {
                      tmp_id = CifCategory[id].keywords[j].NdbInfo[0].NdbTokenId-1;
                      if (tmp_id != HEADER_TOKEN-1 && tmp_id >= 0)
                              ndb_tokenid = tmp_id;
                      break;
                 }
            }
            if (j == CifCategory[id].nkeywords) cols[i] = -1;
            else                                cols[i] = j;
       }
       if (ColumnNames) delete ColumnNames;
       if (ndb_tokenid == -1) {
            delete cols;
            return;
       }

       for (row = 0; row < rowNo; row++) {
            pCard = _CardInfo[ndb_tokenid].add_a_new_card(NdbToken[ndb_tokenid], 0);
            for (col = 0; col < colNo; col++) {
                 j = cols[col];
                 if (j < 0 || j > CifCategory[id].nkeywords) continue;

                 string.Clear();
                 Table->GetCell(string, col, row);
                 if (string == "?" || string == "." ||
                     string == "" || string == "@") continue;

                 ndb_fieldno = CifCategory[id].keywords[j].
                            NdbInfo[0].NdbFieldNo - 1;
                 if (ndb_fieldno < 0) continue;

                 strcpy(value, string.Text());
                 ndb_clean_string(value);
                 pCard->AddElementToField(value, ndb_fieldno);
            }
       }
       delete [] cols;
}

void _FileObj::cifndb_process_refine_ls_restr(ISTable *Table)
{
       int rowNo = Table->GetNumRows();

       int type_index = Table->GetColumnIndex("type");
       if (type_index < 0) return;

       int ave_index  = Table->GetColumnIndex("dev_ideal");
       int std_index  = Table->GetColumnIndex("dev_ideal_target");
       int weight_index = Table->GetColumnIndex("weight");
       int number_index = Table->GetColumnIndex("number");

       int ndb_tokenid, ndb_fieldno, irow, i;
       CifString value;
       for (irow = 0; irow < rowNo; irow++) {
            get_value(value, Table, type_index, irow);
            for (i = 0; i < mxRmsRecords; i++) {
                 if (!strcasecmp(value.Text(), rmsRecords[i].type)) break;
            }
            if (i == mxRmsRecords) continue;

            if (ave_index >= 0 && get_value(value, Table, ave_index, irow)) {
                 ndb_tokenid = rmsRecords[i].tokens[0];
                 ndb_fieldno = rmsRecords[i].fields[0];
                 put_into_card(value.Text(), ndb_tokenid, ndb_fieldno);
            }

            if (std_index >= 0 && get_value(value, Table, std_index, irow)) {
                 ndb_tokenid = rmsRecords[i].tokens[1];
                 ndb_fieldno = rmsRecords[i].fields[1];
                 put_into_card(value.Text(), ndb_tokenid, ndb_fieldno);
            }

            if (weight_index >= 0 && get_value(value, Table, weight_index, irow)) {
                 ndb_tokenid = rmsRecords[i].tokens[2];
                 ndb_fieldno = rmsRecords[i].fields[2];
                 put_into_card(value.Text(), ndb_tokenid, ndb_fieldno);
            }

            if (number_index >= 0 && get_value(value, Table, number_index, irow)) {
                 ndb_tokenid = rmsRecords[i].tokens[3];
                 ndb_fieldno = rmsRecords[i].fields[3];
                 put_into_card(value.Text(), ndb_tokenid, ndb_fieldno);
            }
       }
}

void _FileObj::cifndb_process_matrix(ISTable *Table, const int id)
{
       int colNo = Table->GetNumColumns();
       int rowNo = Table->GetNumRows();

       ReVarCifArray<CifString> *ColumnNames = Table->GetColumnNames();
       int *cols = new int[colNo];
       int ndb_tokenid = -1, i, j, tokenid, fieldno, tmp_id, row;
       CifString value;
       One_Card_Info_Format *pCard1 = NULL, *pCard2 = NULL, *pCard3 = NULL;

       for (i = 0; i < colNo; i++) {
            for (j = 0; j < CifCategory[id].nkeywords; j++) {
                 if (!strcasecmp((*ColumnNames)[i].Text(),
                        CifCategory[id].keywords[j].name)) {
                         tmp_id = CifCategory[id].keywords[j].
                                      NdbInfo[0].NdbTokenId - 1;
                         if (tmp_id != HEADER_TOKEN-1 && tmp_id >= 0)
                              ndb_tokenid = tmp_id;
                      break;
                 }
            }
            if (j == CifCategory[id].nkeywords) cols[i] = -1;
            else                                cols[i] = j;
       }
       if (ndb_tokenid == -1) {
            if (ColumnNames) delete ColumnNames;
            delete cols;
            return;
       }

       for (row = 0; row < rowNo; row++) {
            pCard1 = _CardInfo[ndb_tokenid].add_a_new_card(NdbToken[ndb_tokenid], 0);
            pCard2 = _CardInfo[ndb_tokenid].add_a_new_card(NdbToken[ndb_tokenid], 0);
            pCard3 = _CardInfo[ndb_tokenid].add_a_new_card(NdbToken[ndb_tokenid], 0);
            for (i = 0; i < colNo; i++) {
                 j = cols[i];
                 if (j < 0 || j > CifCategory[id].nkeywords) continue;

                 if (!get_value(value, Table, i, row)) continue;

                 tokenid = CifCategory[id].keywords[j].NdbInfo[0].NdbTokenId-1;
                 fieldno = CifCategory[id].keywords[j].NdbInfo[0].NdbFieldNo-1;
                 if (fieldno < 0 || tokenid != ndb_tokenid) continue;

                 if (strstr((*ColumnNames)[i].Text(), "x[1]") ||
                     strstr((*ColumnNames)[i].Text(), "vector[1]")) {
                      pCard1->AddElementToField(value.Text(), fieldno);
                 } else if (strstr((*ColumnNames)[i].Text(), "x[2]") ||
                            strstr((*ColumnNames)[i].Text(), "vector[2]")) {
                      pCard2->AddElementToField(value.Text(), fieldno);
                 } else if (strstr((*ColumnNames)[i].Text(), "x[3]") ||
                            strstr((*ColumnNames)[i].Text(), "vector[3]")) {
                      pCard3->AddElementToField(value.Text(), fieldno);
                 } else {
                      if (value == "given") value.Copy("1");
                      else if (value == "generate") value.Copy("");
                      pCard1->AddElementToField(value.Text(), fieldno);
                      pCard2->AddElementToField(value.Text(), fieldno);
                      pCard3->AddElementToField(value.Text(), fieldno);
                 }
            }
       }
       if (ColumnNames) delete ColumnNames;
       delete [] cols;
}

void _FileObj::cifndb_process_struct_site_gen(ISTable *Table)
{
       if (_num_mol == 0 || _molecules == NULL) return;

       int rowNo = Table->GetNumRows();

       int site_id_index = Table->GetColumnIndex("site_id");
       int label_comp_id = Table->GetColumnIndex("label_comp_id");
       int label_asym_id = Table->GetColumnIndex("label_asym_id");
       int label_seq_id  = Table->GetColumnIndex("label_seq_id");
       int auth_comp_id  = Table->GetColumnIndex("auth_comp_id");
       int auth_asym_id  = Table->GetColumnIndex("auth_asym_id");
       int auth_seq_id   = Table->GetColumnIndex("auth_seq_id");
       int ndb_ins_code  = Table->GetColumnIndex("ndb_label_ins_code");

       CifString site_id;
       int irow, FstIndex, FstRes;
       for (irow = 0; irow < rowNo; irow++) {
            get_value(site_id, Table, site_id_index, irow);
            if (get_residue_index(_molecules[0], Table, irow, label_comp_id,
                      label_asym_id, label_seq_id, auth_comp_id, auth_asym_id,
                      auth_seq_id, ndb_ins_code, FstIndex, FstRes)) {
                 Insert_a_Site(site_id.Text(), FstIndex, FstRes);
            }
       } 
}

void _FileObj::cifndb_process_rcsb_link(ISTable *Table)
{
       int rowNo, irow, BegIndex, BegRes, EndIndex, EndRes, beg, end;
       int ptnr1_label_alt_id, ptnr1_label_asym_id, ptnr1_label_atom_id;
       int ptnr1_label_comp_id, ptnr1_label_seq_id, ptnr1_label_ins_code;
       int ptnr1_symmetry, ptnr2_label_alt_id, ptnr2_label_asym_id;
       int ptnr2_label_atom_id, ptnr2_label_comp_id, ptnr2_label_seq_id;
       int ptnr2_label_ins_code, ptnr2_symmetry, details;
       CifString comment, atom1, loc1, atom2, loc2, symm1, symm2;

       if (_num_mol == 0 || _molecules == NULL) return;

       rowNo = Table->GetNumRows();
       ptnr1_label_alt_id = Table->GetColumnIndex("ptnr1_label_alt_id");
       ptnr1_label_asym_id = Table->GetColumnIndex("ptnr1_label_asym_id");
       ptnr1_label_atom_id = Table->GetColumnIndex("ptnr1_label_atom_id");
       ptnr1_label_comp_id = Table->GetColumnIndex("ptnr1_label_comp_id");
       ptnr1_label_seq_id = Table->GetColumnIndex("ptnr1_label_seq_id");
       ptnr1_label_ins_code = Table->GetColumnIndex("ptnr1_label_ins_code");
       ptnr1_symmetry = Table->GetColumnIndex("ptnr1_symmetry");
       ptnr2_label_alt_id = Table->GetColumnIndex("ptnr2_label_alt_id");
       ptnr2_label_asym_id = Table->GetColumnIndex("ptnr2_label_asym_id");
       ptnr2_label_atom_id = Table->GetColumnIndex("ptnr2_label_atom_id");
       ptnr2_label_comp_id = Table->GetColumnIndex("ptnr2_label_comp_id");
       ptnr2_label_seq_id = Table->GetColumnIndex("ptnr2_label_seq_id");
       ptnr2_label_ins_code = Table->GetColumnIndex("ptnr2_label_ins_code");
       ptnr2_symmetry = Table->GetColumnIndex("ptnr2_symmetry");
       details = Table->GetColumnIndex("details");

       for (irow = 0; irow < rowNo; irow++) {
            get_value(comment, Table, details, irow);
            get_value(loc1, Table, ptnr1_label_alt_id, irow);
            get_value(atom1, Table, ptnr1_label_atom_id, irow);
            get_value(symm1, Table, ptnr1_symmetry, irow);
            get_value(loc2, Table, ptnr2_label_alt_id, irow);
            get_value(atom2, Table, ptnr2_label_atom_id, irow);
            get_value(symm2, Table, ptnr2_symmetry, irow);
            beg = get_residue_index(_molecules[0], Table, irow, -1, -1, -1,
                   ptnr1_label_comp_id, ptnr1_label_asym_id,
                   ptnr1_label_seq_id, ptnr1_label_ins_code,
                   BegIndex, BegRes);
            end = get_residue_index(_molecules[0], Table, irow, -1, -1, -1,
                   ptnr2_label_comp_id, ptnr2_label_asym_id,
                   ptnr2_label_seq_id, ptnr2_label_ins_code,
                   EndIndex, EndRes);
            if (!beg || !end) continue;
            if (symm1 == "1_555") symm1.Clear();
            if (symm2 == "1_555") symm2.Clear();
            Insert_a_Link(atom1.Text(), loc1.Text(), BegIndex, BegRes,
                          atom2.Text(), loc2.Text(), EndIndex, EndRes,
                          symm1.Text(), symm2.Text(), comment.Text());
       }
}

void _FileObj::cifndb_process_struct_conf(ISTable *Table)
{
       int BegIndex = -1, BegRes = -1, EndIndex = -1, EndRes = -1, beg = -1, end = -1;
       int irow, rowNo, id_index, pdb_index, beg_label_comp_id, beg_label_asym_id;
       int beg_label_seq_id, beg_label_ins_code, beg_auth_comp_id, beg_auth_asym_id;
       int beg_auth_seq_id, end_label_comp_id, end_label_asym_id, end_label_seq_id;
       int end_label_ins_code, end_auth_comp_id, end_auth_asym_id, end_auth_seq_id;
       int class_index, details_index;
       CifString pdb_id, type, comment, class_type;
       _CHAIN *chain = NULL;

       if (_num_mol == 0 || _molecules == NULL) return;

       rowNo = Table->GetNumRows();
       id_index = Table->GetColumnIndex("conf_type_id");
       pdb_index = Table->GetColumnIndex("pdb_id");
       beg_label_comp_id = Table->GetColumnIndex("beg_label_comp_id");
       beg_label_asym_id = Table->GetColumnIndex("beg_label_asym_id");
       beg_label_seq_id = Table->GetColumnIndex("beg_label_seq_id");
       beg_label_ins_code = Table->GetColumnIndex("ndb_beg_label_ins_code_pdb");
       beg_auth_comp_id = Table->GetColumnIndex("beg_auth_comp_id");
       beg_auth_asym_id = Table->GetColumnIndex("beg_auth_asym_id");
       beg_auth_seq_id = Table->GetColumnIndex("beg_auth_seq_id");
       end_label_comp_id = Table->GetColumnIndex("end_label_comp_id");
       end_label_asym_id = Table->GetColumnIndex("end_label_asym_id");
       end_label_seq_id = Table->GetColumnIndex("end_label_seq_id");
       end_label_ins_code = Table->GetColumnIndex("ndb_end_label_ins_code_pdb");
       end_auth_comp_id = Table->GetColumnIndex("end_auth_comp_id");
       end_auth_asym_id = Table->GetColumnIndex("end_auth_asym_id");
       end_auth_seq_id = Table->GetColumnIndex("end_auth_seq_id");
       class_index = Table->GetColumnIndex("ndb_helix_class_pdb");
       details_index = Table->GetColumnIndex("details");

       for (irow = 0; irow < rowNo; irow++) {
            beg = get_residue_index(_molecules[0], Table, irow, beg_label_comp_id,
                  beg_label_asym_id, beg_label_seq_id, beg_auth_comp_id,
                  beg_auth_asym_id, beg_auth_seq_id, beg_label_ins_code,
                  BegIndex, BegRes);
            end = get_residue_index(_molecules[0], Table, irow, end_label_comp_id,
                  end_label_asym_id, end_label_seq_id, end_auth_comp_id,
                  end_auth_asym_id, end_auth_seq_id, end_label_ins_code,
                  EndIndex, EndRes);
            if (beg && end && BegIndex == EndIndex) {
                 chain = _molecules[0]->GetIndexChain(BegIndex);
                 if (chain) {
                      get_value(type, Table, id_index, irow);
                      get_value(pdb_id, Table, pdb_index, irow);
                      get_value(comment, Table, details_index, irow);
                      if (!strcasecmp(type.Text(), "HELX_P")) {
                           get_value(class_type, Table, class_index, irow);
                           chain->Insert_a_Helix(pdb_id.Text(), BegRes, EndRes,
                                 atoi(class_type.Text()), comment.Text());
                      } else if (!strcasecmp(type.Text(), "TURN_P")) {
                           chain->Insert_a_Turn(pdb_id.Text(), BegRes, EndRes, comment.Text());
                      }
                 }
            }
       }
}

void _FileObj::cifndb_process_sheet()
{
       if (_num_mol == 0 || _molecules == NULL) return;

       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "struct_sheet_range");
       if (is_empty_table(Table)) return;

       ISTable *Table1 = _CifObj->GetTablePtr(_StructureId, "rcsb_struct_sheet_hbond");
       ISTable *Table2 = _CifObj->GetTablePtr(_StructureId, "struct_sheet_order");
       ISTable *Table3 = _CifObj->GetTablePtr(_StructureId, "struct_sheet");

       int rowNo = Table->GetNumRows();
       int sheet_id_index = Table->GetColumnIndex("sheet_id");
       int id_index = Table->GetColumnIndex("id");
       int beg_label_comp_id = Table->GetColumnIndex("beg_label_comp_id");
       int beg_label_asym_id = Table->GetColumnIndex("beg_label_asym_id");
       int beg_label_seq_id = Table->GetColumnIndex("beg_label_seq_id");
       int ndb_beg_label_ins_code_pdb = Table->GetColumnIndex("ndb_beg_label_ins_code_pdb");
       int end_label_comp_id = Table->GetColumnIndex("end_label_comp_id");
       int end_label_asym_id = Table->GetColumnIndex("end_label_asym_id");
       int end_label_seq_id = Table->GetColumnIndex("end_label_seq_id");
       int ndb_end_label_ins_code_pdb = Table->GetColumnIndex("ndb_end_label_ins_code_pdb");
       int beg_auth_comp_id = Table->GetColumnIndex("beg_auth_comp_id");
       int beg_auth_asym_id = Table->GetColumnIndex("beg_auth_asym_id");
       int beg_auth_seq_id = Table->GetColumnIndex("beg_auth_seq_id");
       int end_auth_comp_id = Table->GetColumnIndex("end_auth_comp_id");
       int end_auth_asym_id = Table->GetColumnIndex("end_auth_asym_id");
       int end_auth_seq_id = Table->GetColumnIndex("end_auth_seq_id");

       int row, BegIndex, BegRes, EndIndex, EndRes, beg, end, number_strands;
       int sense = 0, curIndex = -1, curRes = -1, prevIndex = -1, prevRes = -1;
       CifString sheet_id, id, prev_id, curAtom, prevAtom;
       prev_id[0] = '\0';
       prev_id.Clear();
       id.Clear();
       for (row = 0; row < rowNo; row++) {
            prev_id.Copy(id);
            sheet_id.Clear();
            id.Clear();
            get_value(sheet_id, Table, sheet_id_index, row);
            get_value(id, Table, id_index, row);

            beg = get_residue_index(_molecules[0], Table, row, beg_label_comp_id,
                  beg_label_asym_id, beg_label_seq_id, beg_auth_comp_id,
                  beg_auth_asym_id, beg_auth_seq_id, ndb_beg_label_ins_code_pdb,
                  BegIndex, BegRes);
            end = get_residue_index(_molecules[0], Table, row, end_label_comp_id,
                  end_label_asym_id, end_label_seq_id, end_auth_comp_id,
                  end_auth_asym_id, end_auth_seq_id, ndb_end_label_ins_code_pdb,
                  EndIndex, EndRes);
            if (beg && end && BegIndex == EndIndex) {
                 sense = 0; curIndex = -1; curRes = -1; prevIndex = -1; prevRes = -1;
                 curAtom.Clear(); prevAtom.Clear();
                 number_strands = get_number_strands(Table3, sheet_id.Text());
                 if (row) {
                      get_hbond(_molecules[0], Table1, BegIndex, sheet_id.Text(), id.Text(),
                         curAtom, curIndex, curRes, prevAtom, prevIndex, prevRes);
                      sense = get_sense(Table2, sheet_id.Text(), prev_id.Text(), id.Text());
                 }
                 Insert_a_Sheet(sheet_id.Text(), number_strands, BegIndex, BegRes, EndIndex,
                       EndRes, sense, curAtom.Text(), curIndex, curRes,
                       prevAtom.Text(), prevIndex, prevRes);
            }
       }
}

void _FileObj::cifndb_process_database_pdb_rev(ISTable *Table)
{
       int rowNo = Table->GetNumRows();
       PDB_REV *pdb_rev = (PDB_REV *) calloc(rowNo, sizeof(PDB_REV));
       int n_rev = rowNo;
       int num_index = Table->GetColumnIndex("num");
       int date_index = Table->GetColumnIndex("date");
       int original_index = Table->GetColumnIndex("date_original");
       int replaces_index = Table->GetColumnIndex("replaces");
       int type_index = Table->GetColumnIndex("mod_type");
       int revised_index[4];
       revised_index[0] = Table->GetColumnIndex("rcsb_record_revised_1");
       revised_index[1] = Table->GetColumnIndex("rcsb_record_revised_2");
       revised_index[2] = Table->GetColumnIndex("rcsb_record_revised_3");
       revised_index[3] = Table->GetColumnIndex("rcsb_record_revised_4");

       int irow, i, len, serial_no, pos;
       CifString cifstring, value;
       char buffer[20], *p;
       One_Card_Info_Format *pCard = NULL;
       for (irow = 0; irow < rowNo; irow++) {
            pdb_rev[irow].num = 0;
            pdb_rev[irow].date[0] = '\0';
            pdb_rev[irow].replaces[0] = '\0';
            pdb_rev[irow].mod_type = 0;
            pdb_rev[irow].record = NULL;
            if (num_index >= 0) {
                 get_value(cifstring, Table, num_index, irow);
                 pdb_rev[irow].num = atoi(cifstring.Text());
            }
            if (date_index >= 0) {
                 get_value(cifstring, Table, date_index, irow);
                 strcpy(pdb_rev[irow].date, cifstring.Text());
            }
            if (replaces_index >= 0) {
                 get_value(cifstring, Table, replaces_index, irow);
                 strcpy(pdb_rev[irow].replaces, cifstring.Text());
            }
            if (type_index >= 0) {
                 get_value(cifstring, Table, type_index, irow);
                 pdb_rev[irow].mod_type = atoi(cifstring.Text());
            }
            if (original_index >= 0 && !pdb_rev[irow].mod_type) {
                 get_value(cifstring, Table, original_index, irow);
                 if (cifstring != "") {
                      pCard = _CardInfo[HEADER_TOKEN-1].Head();
                      if (!pCard) pCard = _CardInfo[HEADER_TOKEN-1].
                           add_a_new_card(NdbToken[HEADER_TOKEN-1], 0);
                      pCard->AddElementToField(cifstring.Text(), 2);
                 }
            }
            for (i = 0; i < 4; i++) {
                 if (revised_index[i] < 0) continue;
                 get_value(cifstring, Table, revised_index[i], irow);
                 if (cifstring == "") continue;

                 if (pdb_rev[irow].record == NULL) {
                      len = strlen(cifstring.Text())+1;
                      pdb_rev[irow].record = (char *) calloc(len, sizeof(char));
                      strcpy(pdb_rev[irow].record, cifstring.Text());
                 } else if (!strstr(pdb_rev[irow].record, cifstring.Text())) {
                      len = strlen(cifstring.Text()) + 2 +
                                strlen(pdb_rev[irow].record);
                      pdb_rev[irow].record = (char *) realloc(pdb_rev[irow].record,
                                            len * sizeof(char));
                      strcat(pdb_rev[irow].record, " ");
                      strcat(pdb_rev[irow].record, cifstring.Text());
                 }
            }
       }

       Table = _CifObj->GetTablePtr(_StructureId, "database_PDB_rev_record");
       if (!is_empty_table(Table)) {
            rowNo = Table->GetNumRows();
            num_index = Table->GetColumnIndex("rev_num");
            type_index = Table->GetColumnIndex("type");
            for (irow = 0; irow < rowNo; irow++) {
                 get_value(cifstring, Table, num_index, irow);
                 get_value(value, Table, type_index, irow);
                 if (cifstring == "" || value == "") continue;

                 for (i = 0; i < n_rev; i++)
                      if (pdb_rev[i].num == atoi(cifstring.Text())) break;
                 if (i < n_rev) {
                      if (pdb_rev[i].record == NULL) {
                           len = strlen(value.Text())+1;
                           pdb_rev[i].record = (char *) calloc(len, sizeof(char));
                           strcpy(pdb_rev[i].record, value.Text());
                      } else if (!strstr(pdb_rev[i].record, value.Text())) {
                           len = strlen(value.Text()) + 2 +
                                     strlen(pdb_rev[i].record);
                           pdb_rev[i].record = (char *) realloc(pdb_rev[i].record,
                                                 len * sizeof(char));
                           strcat(pdb_rev[i].record, " ");
                           strcat(pdb_rev[i].record, value.Text());
                      }
                 }
            }
       }
       if (n_rev > 1) qsort((void *) pdb_rev, n_rev, sizeof(PDB_REV), pdb_rev_compare);

       for (i = 0; i < n_rev; i++) {
            if (!pdb_rev[i].num) continue;
            pCard = _CardInfo[REVDAT_TOKEN-1].
                        add_a_new_card(NdbToken[REVDAT_TOKEN-1], 0);
            pCard->AddElementToField(pdb_rev[i].num, 1);
            pCard->AddElementToField(pdb_rev[i].date, 3);
            pCard->AddElementToField(pdb_rev[i].replaces, 4);
            pCard->AddElementToField(pdb_rev[i].mod_type, 5);
            p = pdb_rev[i].record;
            serial_no = 2;
            pos = 6;
            while (p) {
                 if (pos > 9) {
                      pCard = _CardInfo[REVDAT_TOKEN-1].
                              add_a_new_card(NdbToken[REVDAT_TOKEN-1], 0);
                      pCard->AddElementToField(pdb_rev[i].num, 1);
                      pCard->AddElementToField(serial_no, 2);
                      pCard->AddElementToField(pdb_rev[i].mod_type, 5);
                      serial_no++;
                      pos = 6;
                 }
                 p = get_word(p, buffer);
                 pCard->AddElementToField(buffer, pos);
                 pos++;
            }
       }
       for (i = 0; i < n_rev; i++) {
            if (pdb_rev[i].record) free ((void *) pdb_rev[i].record);
       }
       free ((void *) pdb_rev);
}

void _FileObj::cifndb_process_struct_conn(ISTable *Table)
{
       if (_num_mol == 0 || _molecules == NULL) return;

       int rowNo = Table->GetNumRows();

       int id_index = Table->GetColumnIndex("conn_type_id");
       int pdb_index = Table->GetColumnIndex("ndb_PDB_id");
       int name_index = Table->GetColumnIndex("ndb_ptnr1_mod_name");
       int replaced_index = Table->GetColumnIndex("ndb_ptnr1_replaced_atom");
       int standard_index = Table->GetColumnIndex("ndb_ptnr1_standard_comp_id");
       int ptnr1_label_comp_id = Table->GetColumnIndex("ptnr1_label_comp_id");
       int ptnr1_label_asym_id = Table->GetColumnIndex("ptnr1_label_asym_id");
       int ptnr1_label_seq_id = Table->GetColumnIndex("ptnr1_label_seq_id");
       int ptnr1_label_atom_id = Table->GetColumnIndex("ptnr1_label_atom_id");
       int ndb_ptnr1_label_alt_loc = Table->GetColumnIndex("ndb_ptnr1_label_alt_loc");
       int ptnr1_label_ins_code = Table->GetColumnIndex("ndb_ptnr1_label_ins_code");
       int ptnr1_symmetry = Table->GetColumnIndex("ptnr1_symmetry");
       int ptnr2_label_comp_id = Table->GetColumnIndex("ptnr2_label_comp_id");
       int ptnr2_label_asym_id = Table->GetColumnIndex("ptnr2_label_asym_id");
       int ptnr2_label_seq_id = Table->GetColumnIndex("ptnr2_label_seq_id");
       int ptnr2_label_atom_id = Table->GetColumnIndex("ptnr2_label_atom_id");
       int ndb_ptnr2_label_alt_loc = Table->GetColumnIndex("ndb_ptnr2_label_alt_loc");
       int ptnr2_label_ins_code = Table->GetColumnIndex("ndb_ptnr2_label_ins_code");
       int ptnr2_symmetry = Table->GetColumnIndex("ptnr2_symmetry");
       int ptnr1_auth_comp_id = Table->GetColumnIndex("ptnr1_auth_comp_id");
       int ptnr1_auth_asym_id = Table->GetColumnIndex("ptnr1_auth_asym_id");
       int ptnr1_auth_seq_id = Table->GetColumnIndex("ptnr1_auth_seq_id");
       int ptnr2_auth_comp_id = Table->GetColumnIndex("ptnr2_auth_comp_id");
       int ptnr2_auth_asym_id = Table->GetColumnIndex("ptnr2_auth_asym_id");
       int ptnr2_auth_seq_id = Table->GetColumnIndex("ptnr2_auth_seq_id");
       int ndb_ptnr3_label_atom_id = Table->GetColumnIndex("ndb_ptnr3_label_atom_id");
       int ndb_ptnr3_label_seq_id = Table->GetColumnIndex("ndb_ptnr3_label_seq_id");
       int ndb_ptnr3_label_asym_id = Table->GetColumnIndex("ndb_ptnr3_label_asym_id");
       int ndb_ptnr3_label_alt_loc = Table->GetColumnIndex("ndb_ptnr3_label_alt_loc");
       int details_index = Table->GetColumnIndex("details");

       int irow, BegIndex, BegRes, EndIndex, EndRes, beg, end, H_Index, H_Res;
       CifString type, atom1, loc1, atom2, loc2, symm1, symm2, comment, name;
       CifString Replaced_Atom, atom3, loc3, standard, pdb_id, sugar;
       CifString seq_id1, seq_id2, seq_id3, chain_id1, chain_id2, chain_id3;
       for (irow = 0; irow < rowNo; irow++) {
            get_value(type, Table, id_index, irow);
            if (type == "") continue;

            get_value(atom1, Table, ptnr1_label_atom_id, irow);
            get_value(loc1, Table, ndb_ptnr1_label_alt_loc, irow);
            get_value(symm1, Table, ptnr1_symmetry, irow);
            get_value(atom2, Table, ptnr2_label_atom_id, irow);
            get_value(loc2, Table, ndb_ptnr2_label_alt_loc, irow);
            get_value(symm2, Table, ptnr2_symmetry, irow);
            get_value(comment, Table, details_index, irow);
            get_value(name, Table, name_index, irow);
            get_value(Replaced_Atom, Table, replaced_index, irow);
            if (symm1 == "1_555") symm1.Clear();
            if (symm2 == "1_555") symm2.Clear();

            beg = get_residue_index(_molecules[0], Table, irow, ptnr1_label_comp_id,
                  ptnr1_label_asym_id, ptnr1_label_seq_id, ptnr1_auth_comp_id,
                  ptnr1_auth_asym_id, ptnr1_auth_seq_id, ptnr1_label_ins_code,
                  BegIndex, BegRes);
            end = get_residue_index(_molecules[0], Table, irow, ptnr2_label_comp_id,
                  ptnr2_label_asym_id, ptnr2_label_seq_id, ptnr2_auth_comp_id,
                  ptnr2_auth_asym_id, ptnr2_auth_seq_id, ptnr2_label_ins_code,
                  EndIndex, EndRes);
            if (!beg) continue;

            if (!strcasecmp(type.Text(), "hydrog")) {
                 if (end) {
                      if (!strcasecmp(comment.Text(), "WATSON-CRICK") ||
                          !strcasecmp(comment.Text(), "REVERSED WATSON-CRICK") ||
                          !strcasecmp(comment.Text(), "HOOGSTEEN") ||
                          !strcasecmp(comment.Text(), "REVERSED HOOGSTEEN") ||
                          strstr(comment.Text(), "PAIR") || strstr(comment.Text(), "pair")) {
                      } else {
                           seq_id3.Copy("ZZZ"); chain_id3.Copy("ZZZ");
                           get_value(atom3, Table, ndb_ptnr3_label_atom_id, irow);
                           get_value(loc3, Table, ndb_ptnr3_label_alt_loc, irow);
                           get_value(seq_id1, Table, ptnr1_label_seq_id, irow);
                           get_value(chain_id1, Table, ptnr1_label_asym_id, irow);
                           get_value(seq_id2, Table, ptnr2_label_seq_id, irow);
                           get_value(chain_id2, Table, ptnr2_label_asym_id, irow);
                           if (ndb_ptnr3_label_seq_id >= 0)
                                get_value(seq_id3, Table, ndb_ptnr3_label_seq_id, irow);
                           if (ndb_ptnr3_label_asym_id >= 0)
                                get_value(chain_id3, Table, ndb_ptnr3_label_asym_id, irow);
                           H_Index = -1; H_Res = -1;
                           if (seq_id1 == seq_id3 && chain_id1 == chain_id3) {
                                H_Index = BegIndex; H_Res = BegRes;
                           } else if (seq_id2 == seq_id3 && chain_id2 == chain_id3) {
                                H_Index = EndIndex; H_Res = EndRes;
                           }
                           Insert_a_Hydbnd(atom1.Text(), loc1.Text(), BegIndex, BegRes,
                                 atom3.Text(), loc3.Text(), H_Index, H_Res, atom2.Text(),
                                 loc2.Text(), EndIndex, EndRes, symm1.Text(), symm2.Text());
                      }
                 }
            } else if (!strcasecmp(type.Text(), "hydbnd")) {
                 if (end) {
                      seq_id3.Copy("ZZZ"); chain_id3.Copy("ZZZ");
                      get_value(atom3, Table, ndb_ptnr3_label_atom_id, irow);
                      get_value(loc3, Table, ndb_ptnr3_label_alt_loc, irow);
                      get_value(seq_id1, Table, ptnr1_label_seq_id, irow);
                      get_value(chain_id1, Table, ptnr1_label_asym_id, irow);
                      get_value(seq_id2, Table, ptnr2_label_seq_id, irow);
                      get_value(chain_id2, Table, ptnr2_label_asym_id, irow);
                      if (ndb_ptnr3_label_seq_id >= 0)
                           get_value(seq_id3, Table, ndb_ptnr3_label_seq_id, irow);
                      if (ndb_ptnr3_label_asym_id >= 0)
                           get_value(chain_id3, Table, ndb_ptnr3_label_asym_id, irow);
                      H_Index = -1; H_Res = -1;
                      if (seq_id1 == seq_id3 && chain_id1 == chain_id3) {
                           H_Index = BegIndex; H_Res = BegRes;
                      } else if (seq_id2 == seq_id3 && chain_id2 == chain_id3) {
                           H_Index = EndIndex; H_Res = EndRes;
                      }
                      Insert_a_Hydbnd(atom1.Text(), loc1.Text(), BegIndex, BegRes,
                             atom3.Text(), loc3.Text(), H_Index, H_Res, atom2.Text(),
                             loc2.Text(), EndIndex, EndRes, symm1.Text(), symm2.Text());
                 }
            } else if (!strcasecmp(type.Text(), "link") || !strcasecmp(type.Text(), "covale")) {
                 if (end) Insert_a_Link(atom1.Text(), loc1.Text(), BegIndex, BegRes, atom2.Text(),
                       loc2.Text(), EndIndex, EndRes, symm1.Text(), symm2.Text(), comment.Text());
            } else if (!strcasecmp(type.Text(), "modres")) {
                 get_value(pdb_id, Table, pdb_index, irow);
                 get_value(standard, Table, standard_index, irow);
                 Insert_a_Modres(pdb_id.Text(), BegIndex, BegRes, standard.Text(), comment.Text());
            } else if (!strcasecmp(type.Text(), "sltbrg") || !strcasecmp(type.Text(), "saltbr")) {
                 if (end) Insert_a_Sltbrg(atom1.Text(), loc1.Text(), BegIndex, BegRes, atom2.Text(),
                        loc2.Text(), EndIndex, EndRes, symm1.Text(), symm2.Text());
            } else if (!strcasecmp(type.Text(), "disulf")) {
                 if (end) {
                      Insert_a_SSbond(BegIndex, BegRes, EndIndex, EndRes, symm1.Text(), symm2.Text());
                 }
            }
       }
}

void _FileObj::put_into_card(char *value, const int tokenid, const int fieldno)
{
       if (tokenid < 0 || tokenid >= MxTokens || fieldno < 0 ||
           fieldno >= NdbToken[tokenid].NumField) return;

       One_Card_Info_Format *pCard = _CardInfo[tokenid].Tail();
       if (pCard == NULL || strcmp(pCard->Field(fieldno), ""))
            pCard = _CardInfo[tokenid].add_a_new_card(NdbToken[tokenid], 0);
       pCard->AddElementToField(value, fieldno);
}

void _FileObj::post_processing_from_cif_to_ndb()
{
       int Auth_Token = find_jrnl_id("AUTH");
       if (_JrnlCardInfo[Auth_Token].Head() && _CardInfo[AUTHOR_TOKEN-1].Head() == NULL && 
           strcmp(_JrnlCardInfo[Auth_Token].Head()->Field(4), "")) {
            One_Card_Info_Format *pHet =
                _CardInfo[AUTHOR_TOKEN-1].add_a_new_card(NdbToken[AUTHOR_TOKEN-1], 0);
            pHet->AddElementToField(_JrnlCardInfo[Auth_Token].Head()->Field(4), 2);
       }

       _JrnlCardInfo[Auth_Token].reorder_cards(1, INTEGER_COMPARE, ASCENT);
       Auth_Token = find_jrnl_id("TITL");
       _JrnlCardInfo[Auth_Token].reorder_cards(1, INTEGER_COMPARE, ASCENT);
       Auth_Token = find_jrnl_id("EDIT");
       _JrnlCardInfo[Auth_Token].reorder_cards(1, INTEGER_COMPARE, ASCENT);
       Auth_Token = find_jrnl_id("REF");
       _JrnlCardInfo[Auth_Token].reorder_cards(1, INTEGER_COMPARE, ASCENT);
       Auth_Token = find_jrnl_id("PUBL");
       _JrnlCardInfo[Auth_Token].reorder_cards(1, INTEGER_COMPARE, ASCENT);
       Auth_Token = find_jrnl_id("REFN");
       _JrnlCardInfo[Auth_Token].reorder_cards(1, INTEGER_COMPARE, ASCENT);

       Auth_Token = find_card_id("TITLE");
       One_Card_Info_Format *pCard = _CardInfo[Auth_Token].Head();
       while (pCard) {
            pCard->CleanStringDash(2);
            pCard = pCard->Next();
       }

       if ((_CardInfo[STATUS_TOKEN-1].Head() == NULL ||
           !strcmp(_CardInfo[STATUS_TOKEN-1].Head()->Field(6), "")) &&
            _CardInfo[HEADER_TOKEN-1].Head() &&
            strcmp(_CardInfo[HEADER_TOKEN-1].Head()->Field(2), "")) {
             One_Card_Info_Format *pHet = _CardInfo[STATUS_TOKEN-1].Head();
             if (pHet == NULL) pHet = _CardInfo[STATUS_TOKEN-1].
                         add_a_new_card(NdbToken[STATUS_TOKEN-1], 0);
             pHet->AddElementToField(_CardInfo[HEADER_TOKEN-1].Head()->Field(2), 6);
       }

       pCard = _CardInfo[PROCES_TOKEN-1].Head();
       if (pCard == NULL) pCard = _CardInfo[PROCES_TOKEN-1].
                       add_a_new_card(NdbToken[PROCES_TOKEN-1], 0);
       if (!strcmp(pCard->Field(5), "")) pCard->AddElementToField("RCSB", 5);
 
       _CardInfo[RADIAT_TOKEN-1].reorder_cards(1, INTEGER_COMPARE, ASCENT);

       pCard = _CardInfo[REFMET_TOKEN-1].Head();
       if (pCard && (!strncasecmp(pCard->Field(3), "CNS", 3) ||
           !strncasecmp(pCard->Field(3), "X-PLOR", 6))) {
            pCard = _CardInfo[RFACTR_TOKEN-1].Head();
            if (pCard) {
                 if (!strcmp(pCard->Field(10), "") && strcmp(pCard->Field(7), ""))
                      pCard->AddElementToField(pCard->Field(7), 10);
                 else if (strcmp(pCard->Field(10), "") && !strcmp(pCard->Field(7), ""))
                      pCard->AddElementToField(pCard->Field(10), 7);
            }
       }
       Auth_Token = find_card_id("CAVEAT");
       pCard = _CardInfo[Auth_Token].Head();
       while (pCard) {
            pCard->AddElementToField(_CardInfo[PDBFIL_TOKEN-1].Head()->Field(1), 2);
            pCard = pCard->Next();
       }

       delete_null_cards();
}

void _FileObj::get_het_card()
{
       if (_num_mol == 0 || _molecules == NULL) return;

       if (_het) delete [] _het;
       _het = NULL; _n_het = 0;
       One_Card_Info_Format *pCard = _CardInfo[HET_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[HET_TOKEN-1].delete_a_card(pCard);
       pCard = _CardInfo[FORMUL_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[FORMUL_TOKEN-1].delete_a_card(pCard);
       pCard = _CardInfo[HETNAM_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[HETNAM_TOKEN-1].delete_a_card(pCard);
       pCard = _CardInfo[HETSYN_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[HETSYN_TOKEN-1].delete_a_card(pCard);

       char PDB_ID[15];
       PDB_ID[0] = '\0';
       pCard = _CardInfo[PDBFIL_TOKEN-1].Head();
       if (pCard) strcpy(PDB_ID, pCard->Field(1));

       int i = 0, count = 0, serialNo = 0, num = 0;
       _RESIDUE *residue = NULL;
       _ATOM *atom = NULL;
       ConnectFormat *drug = NULL;
       _CHAIN *chain = _molecules[0]->GetFirstChain();
       while (chain) {
            // serialNo++;
            count++;
            if (chain->chain_type() == HETAS_TOKEN &&
                (!strcmp(chain->SeqRes(0)->Field[0], "HOH") ||
                 !strcmp(chain->SeqRes(0)->Field[0], "DOD"))) {
                 if (!strcmp(chain->SeqRes(0)->Field[0], "HOH"))
                      insert_formul_card("HOH", serialNo, count, chain->ResidueNumbers());
                 else insert_formul_card("DOD", serialNo, count, chain->ResidueNumbers());
                 chain = _molecules[0]->GetNextChain();
                 continue;
            }

            for (i = 0; i < chain->SeqLen(); i++) {
                 residue = chain->SeqRes(i)->Residue;
                 if (!residue) continue;

                 atom = residue->GetFirstAtom();
                 if (!is_a_standard_residue(residue->ResName())) {
                      insert_formul_card(residue->ResName(), serialNo, count, 1);
                      insert_het_card(chain, residue->ResName(),
                            atom->pdb_resnum(), atom->ins_code(),
                            residue->AtomNumbers());
                      if ((chain->chain_type() == ATOMP_TOKEN ||
                           chain->chain_type() == ATOMN_TOKEN) &&
                           strcmp(residue->ResName(), "ACE") &&
                           strcmp(residue->ResName(), "NH2") &&
                           strcmp(residue->ResName(), "NH3") &&
                           strcmp(residue->ResName(), "CH2")) {
                           drug = find_drug(residue->ResName());
                           if (drug) {
                                if (strlen(drug->chemical_name()) < 40)
                                     Insert_a_Modres(PDB_ID, chain->index(),
                                          chain->SeqRes(i)->ResIndex,
                                          drug->parent_residue(),
                                          drug->chemical_name());
                                else Insert_a_Modres(PDB_ID, chain->index(),
                                          chain->SeqRes(i)->ResIndex,
                                          drug->parent_residue(), "");
                           }
                      }
                 } else if (is_a_standard_residue(residue->ResName()) &&
                            chain->chain_type() != ATOMP_TOKEN &&
                            chain->chain_type() != ATOMN_TOKEN) {
                      insert_formul_card(residue->ResName(), serialNo, count, 1);
                      insert_het_card(chain, residue->ResName(),
                            atom->pdb_resnum(),
                            atom->ins_code(), residue->AtomNumbers());
                 }
            }
            chain = _molecules[0]->GetNextChain();
       }
 
       serialNo = 0;
       char buff[50], buffer[20];
       One_Card_Info_Format *pHetNam = NULL, *pHetSyn = NULL;
       pCard = _CardInfo[FORMUL_TOKEN-1].Head();
       while (pCard) {
            num = atoi(pCard->Field(5));
            if (!strcmp(pCard->Field(2), "HOH")) {
                 pCard->AddElementToField("*", 4);
                 if (num > 1) {
                      sprintf(buffer, "%d(H2 O1)", num);
                      pCard->AddElementToField(buffer, 5);
                 } else pCard->AddElementToField("(H2 O1)", 5);
            } else {
                 drug = find_drug(pCard->Field(2));
                 if (drug) {
                      buff[0] = '\0';
                      if (num > 1) {
                           if (drug->formal_charge() &&
                               strcmp(drug->formal_charge(), "") &&
                               strcmp(drug->formal_charge(), "0"))
                                sprintf(buff, "%d(%s %s)", num, drug->formula(),
                                                 drug->formal_charge());
                           else sprintf(buff, "%d(%s)", num, drug->formula());
                      } else {
                           if (drug->formal_charge() &&
                               strcmp(drug->formal_charge(), "") &&
                               strcmp(drug->formal_charge(), "0"))
                                sprintf(buff, "%s %s", drug->formula(),
                                                drug->formal_charge());
                           else strcpy(buff, drug->formula());
                      }
                      pCard->AddElementToField(buff, 5);

                      if (drug->chemical_name() &&
                          strcmp(drug->chemical_name(), "")) {
                           pHetNam = _CardInfo[HETNAM_TOKEN-1].
                                      add_a_new_card(NdbToken[HETNAM_TOKEN-1], 0);
                           pHetNam->AddElementToField(pCard->Field(2), 2);
                           pHetNam->AddElementToField(drug->chemical_name(), 3);
                           pHetNam->AddElementToField(++serialNo, 4);
                      }

                      if (drug->synonym() && strcmp(drug->synonym(), "")) {
                           pHetSyn = _CardInfo[HETSYN_TOKEN-1].
                                      add_a_new_card(NdbToken[HETSYN_TOKEN-1], 0);
                           pHetSyn->AddElementToField(pCard->Field(2), 2);
                           pHetSyn->AddElementToField(drug->synonym(), 3);
                      }
                 } else pCard->AddElementToField("", 5);
            }
            pCard = pCard->Next();
       }
}

void _FileObj::insert_formul_card(const char *hetID, int &serialNo, const int count,  const int num)
{
       One_Card_Info_Format *pCard = _CardInfo[FORMUL_TOKEN-1].Head();
       while (pCard) {
            if (!strcmp(pCard->Field(2), hetID)) break;
            pCard = pCard->Next();
       }
       if (pCard) {
            int n = atoi(pCard->Field(5)) + num;
            pCard->AddElementToField(n, 5);
       } else {
            if (!serialNo)
                 serialNo = count;
            else serialNo++;
            pCard = _CardInfo[FORMUL_TOKEN-1].
                 add_a_new_card(NdbToken[FORMUL_TOKEN-1], 0);
            pCard->AddElementToField(serialNo, 1);
            pCard->AddElementToField(hetID, 2);
            pCard->AddElementToField(num, 5);
       }
}

void _FileObj::insert_het_card(_CHAIN *chain, const char *resname,
       const int resnum, const char *ins_code, const int atmnum)
{
       if (!strcmp(resname, "DUM") || !strcmp(resname, "UNK") ||
           !strcmp(resname, "N")) return;
       _FIELD* initRes = chain->find_pdb_Residue(resname, resnum, ins_code);
       if (initRes) {
            int FstIndex = chain->index();
            int FstRes = initRes->ResIndex;
            Insert_a_Het(FstIndex, FstRes, atmnum);
       }
}

static int get_number_strands(ISTable *Table, const char *sheet_id)
{
       int  id_index, num_index, rowNo, irow;
       CifString id, number;
       if (is_empty_table(Table)) return 1;

       id_index = Table->GetColumnIndex("id");
       num_index = Table->GetColumnIndex("number_strands");
       if (id_index < 0 || num_index < 0) return 1;
       rowNo = Table->GetNumRows();
       for (irow = 0; irow < rowNo; irow++) {
            get_value(id, Table, id_index, irow);
            get_value(number, Table, num_index, irow);
            if (!strcasecmp(sheet_id, id.Text())) {
                 return (atoi(number.Text()));
            }
       }
       return 1;
}

static int get_sense(ISTable *Table, const char *sheet_id, const char *id_1, const char *id_2)
{
       int id_index, range_id_1, range_id_2, sense_index, rowNo, irow;
       char sense[50];
       CifString id, range_1, range_2, cifstring;

       if (is_empty_table(Table)) return 0;
       id_index = Table->GetColumnIndex("sheet_id");
       range_id_1 = Table->GetColumnIndex("range_id_1");
       range_id_2 = Table->GetColumnIndex("range_id_2");
       sense_index  = Table->GetColumnIndex("sense");
       if (id_index < 0 || range_id_1 < 0 || range_id_2 < 0 || sense_index < 0) return 0;
       rowNo = Table->GetNumRows();
       for (irow = 0; irow < rowNo; irow++) {
            get_value(id, Table, id_index, irow);
            get_value(range_1, Table, range_id_1, irow);
            get_value(range_2, Table, range_id_2, irow);
            get_value(cifstring, Table, sense_index, irow);
            if (!strcasecmp(sheet_id, id.Text()) && !strcasecmp(id_1, range_1.Text()) &&
                !strcasecmp(id_2, range_2.Text())) {
                 strcpy(sense, cifstring.Text());
                 ndb_clean_string(sense);
                 if (sense[0] == '-' || isdigit(sense[0])) return (atoi(sense));
                 else if (!strcasecmp(sense, "anti-parallel")) return (-1);
                 else if (!strcasecmp(sense, "parallel")) return (1);
            }
       }
       return 0;
}

static void get_hbond(_MOLECULE *mol, ISTable *Table, const int chain_index,
            const char *sheet_id, const char *range_id, CifString &curAtom, int &curIndex,
            int &curRes, CifString &prevAtom, int &prevIndex, int &prevRes)
{
       if (is_empty_table(Table)) return;
       int id_index = Table->GetColumnIndex("sheet_id");
       int range_id_2 = Table->GetColumnIndex("range_id_2");
       int beg_label_atom_id = Table->GetColumnIndex("range_2_label_atom_id");
       int beg_label_comp_id = Table->GetColumnIndex("range_2_label_comp_id");
       int beg_label_asym_id = Table->GetColumnIndex("range_2_label_asym_id");
       int beg_label_seq_id = Table->GetColumnIndex("range_2_label_seq_id");
       int beg_label_ins_code = Table->GetColumnIndex("range_2_PDB_ins_code");
       int beg_auth_comp_id = Table->GetColumnIndex("range_2_auth_comp_id");
       int beg_auth_asym_id = Table->GetColumnIndex("range_2_auth_asym_id");
       int beg_auth_seq_id = Table->GetColumnIndex("range_2_auth_seq_id");
       int end_label_atom_id = Table->GetColumnIndex("range_1_label_atom_id");
       int end_label_comp_id = Table->GetColumnIndex("range_1_label_comp_id");
       int end_label_asym_id = Table->GetColumnIndex("range_1_label_asym_id");
       int end_label_seq_id = Table->GetColumnIndex("range_1_label_seq_id");
       int end_label_ins_code = Table->GetColumnIndex("range_1_PDB_ins_code");
       int end_auth_comp_id = Table->GetColumnIndex("range_1_auth_comp_id");
       int end_auth_asym_id = Table->GetColumnIndex("range_1_auth_asym_id");
       int end_auth_seq_id = Table->GetColumnIndex("range_1_auth_seq_id");
       int rowNo = Table->GetNumRows();

       int irow, BegIndex, BegRes, EndIndex, EndRes, beg, end;
       CifString id, range_2;
       for (irow = 0; irow < rowNo; irow++) {
            get_value(id, Table, id_index, irow);
            get_value(range_2, Table, range_id_2, irow);
            if (!strcasecmp(sheet_id, id.Text()) && !strcasecmp(range_id, range_2.Text())) {
                 beg = get_residue_index(mol, Table, irow, beg_label_comp_id,
                       beg_label_asym_id, beg_label_seq_id, beg_auth_comp_id,
                       beg_auth_asym_id, beg_auth_seq_id, beg_label_ins_code,
                       BegIndex, BegRes);
                 end = get_residue_index(mol, Table, irow, end_label_comp_id,
                       end_label_asym_id, end_label_seq_id, end_auth_comp_id,
                       end_auth_asym_id, end_auth_seq_id, end_label_ins_code,
                       EndIndex, EndRes);
                 if (beg && end && (BegIndex == chain_index || EndIndex == chain_index)) {
                      curIndex = BegIndex; curRes = BegRes;
                      prevIndex = EndIndex; prevRes = EndRes;
                      get_value(curAtom, Table, beg_label_atom_id, irow);
                      get_value(prevAtom, Table, end_label_atom_id, irow);
                 }
                 return;
            }
       }
}

static void cif_to_ndb_find_serial_no(ISTable *Table, const int id, int irow,
         int tokenid, int *idfield, char idvalue[10][82], char SerialNo[10][82])
{
       int i, j, k, str_len;
       CifString value;
       for (j = 0; j < 10; j++) {
            SerialNo[j][0] = '\0';
            idvalue[j][0] = '\0';
       }

       if (CifCategory[id].nCifKeyField) {
            for (i = 0; i < CifCategory[id].nCifKeyField; i++) {
                 j =  CifCategory[id].CifKeyField[i];
                 get_value(value, Table, idfield[i], irow);
                 if (tokenid == JRNL_TOKEN-1) {
                      if (value == "primary")
                           strcpy(SerialNo[i], "0");
                      else if (value != "")
                           strcpy(SerialNo[i], value.Text());
                      else sprintf(SerialNo[i], "%d", irow);
                 } else if (value != "") {
                      strcpy(SerialNo[i], value.Text());
                      for (k = 0; k < CifCategory[id].keywords[j].numNdbInfo; k++) {
                           str_len = strlen(CifCategory[id].keywords[j].NdbInfo[k].
                                   JrnlTokenName_Type);
                           if (str_len != 0 && (int) strlen(value.Text()) >= str_len &&
                              !strncmp(value.Text(), CifCategory[id].keywords[j].NdbInfo[k].
                                   JrnlTokenName_Type, str_len)) {
                                strcpy(SerialNo[i], &value.Text()[str_len]);
                                strcpy(idvalue[i], CifCategory[id].keywords[j].
                                      NdbInfo[k].JrnlTokenName_Type);
                                break;
                           }
                      }
                 } else sprintf(SerialNo[i], "%d", irow + 1);
            }
       } else sprintf(SerialNo[0], "%d", irow + 1);
}

#ifdef SUN_OS
#ifdef __cplusplus
extern "C" {
#endif
#endif
static int pdb_rev_compare(const void *a, const void *b)
{
       return (((PDB_REV *) b)->num - ((PDB_REV *) a)->num);
}
#ifdef SUN_OS
#ifdef __cplusplus
}
#endif
#endif

static int get_residue_index(_MOLECULE *mol, ISTable *Table, const int irow,
                const int ndb_comp_id, const int ndb_asym_id, const int ndb_seq_id,
                const int pdb_comp_id, const int pdb_asym_id, const int pdb_seq_id,
                const int ins_code_id, int &FstIndex, int &FstRes)
{
       FstIndex = -1; FstRes = -1;

       CifString res_name, chain_id, ins_code, value;
       int seq_num = -99999;
       get_value(res_name, Table, ndb_comp_id, irow);
       get_value(chain_id, Table, ndb_asym_id, irow);
       if (get_value(value, Table, ndb_seq_id, irow))
            seq_num = atoi(value.Text());
       get_value(ins_code, Table, ins_code_id, irow);

       if (chain_id != "" && res_name != "" && seq_num != -99999 &&
           mol->find_ndb_residue(chain_id.Text(), res_name.Text(), seq_num,
           ins_code.Text(), FstIndex, FstRes)) return 1;

       get_value(res_name, Table, pdb_comp_id, irow);
       get_value(chain_id, Table, pdb_asym_id, irow);
       if (get_value(value, Table, pdb_seq_id, irow))
            seq_num = atoi(value.Text());

       if (pdb_comp_id >= 0 && pdb_asym_id >= 0 && pdb_seq_id >= 0 &&
           mol->find_pdb_residue(chain_id.Text(), res_name.Text(), seq_num,
           ins_code.Text(), FstIndex, FstRes)) return 1;

       return 0;
}
