/*
FILE:     util.C
*/
/*
VERSION:  2.053
*/
/*
DATE:     1/24/2007
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2007 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "ndb_misclib.h"
#include "partime.h"
#include "erf_connect.h"
#include "erf_iolib.h"
#include "erf_util.h"
#include "util.h"

int connect_init(const char *path)
{
       ConnectDictionary = new Connect_Dic();
       if (!ConnectDictionary) return -1;
       if (ConnectDictionary->OpenFile(path) < 0) return -1;
       return 0;
}

void connect_close()
{
       ConnectDictionary->CloseFile();
       delete ConnectDictionary;
       ConnectDictionary = NULL;
}

ConnectFormat *find_drug(const char *drugName)
{
       ConnectFormat *drug = ConnectDictionary->find_drug(drugName);
       if (drug && drug->ndbAtomType() < 0) {
            int card_id = find_card_id(drug->AtomType()) + 1;
            drug->set_ndbAtomType(card_id);
       }
       return drug;
}

int find_residue_type(const char *residueName)
{
       ConnectFormat *drug = find_drug(residueName);
       if (drug) return (drug->ndbAtomType());
       else {
            ndb_log_message(NDB_MSG_ERR, "Unrecognized residue name %s\n",
                   residueName);
            return (HETAIN_TOKEN);
       }
}

void get_field_format(const int card_id, const int field_no, char *FormatString,
         const int IsJRNL)
{
       if (IsJRNL == FALSE) {
            if (NdbToken[card_id].FieldList[field_no].FieldType > 2) {
                 if ((NdbToken[card_id].FieldList[field_no].FieldType % 2) == 1)
                      sprintf(FormatString, "%%-%ds",
                              NdbToken[card_id].FieldList[field_no].FieldWidth);
                 else sprintf(FormatString, "%%%ds",
                              NdbToken[card_id].FieldList[field_no].FieldWidth);
            } else sprintf(FormatString, "%%%d.%dlf",
                     NdbToken[card_id].FieldList[field_no].FieldWidth,
                     NdbToken[card_id].FieldList[field_no].FieldPrec);
       } else {
            if (JrnlToken[card_id].FieldList[field_no].FieldType > 2) {
                 if ((JrnlToken[card_id].FieldList[field_no].FieldType % 2) == 1)
                      sprintf(FormatString, "%%-%ds",
                              JrnlToken[card_id].FieldList[field_no].FieldWidth);
                 else sprintf(FormatString, "%%%ds",
                              JrnlToken[card_id].FieldList[field_no].FieldWidth);
            } else sprintf(FormatString, "%%%dd",
                     JrnlToken[card_id].FieldList[field_no].FieldWidth);
       }
}

char *get_max_length_word(char *string, char *words, const int max_length)
{
       int i, str_len, length;
       char *p;

       str_len = strlen(string);
       if (str_len <= max_length) {
            strcpy(words, string);
            return NULL;
       }

       words[0] = '\0';
       length = 0;
       for (i = 0; i < str_len; i++) {
            if (string[i] == ' ') {
                if (i >= max_length) break;
                length = i;
            } else if (!isalnum(string[i]) && string[i] != '.' &&
                  string[i] != '(' && string[i] != '[') {
                if ((i + 1) >= max_length) break;
                length = i + 1;
            } else if (isalpha(string[i]) && (string[i+1] == '-' ||
                  string[i+1] == '(' || string[i+1] == '[')) {
                if ((i + 1) >= max_length) break;
                length = i + 1;
            }
       }

       for (i = 0; i < length; i++) words[i] = string[i];
       words[length] = '\0';

       p = string + length;
       ndb_clean_string(p);
       if (p[0] == '\0') return NULL;
       else              return p;
}

int is_a_standard_residue(char *resname)
{
       static char standard_residue_list[30][5] = {"A", "C", "G", "I", "T", "U", "N", "ALA",
             "ARG", "ASN", "ASP", "ASX", "CYS", "GLN", "GLU", "GLX", "GLY", "HIS", "ILE",
             "LEU", "LYS", "MET", "PHE", "PRO", "SER", "THR", "TRP", "TYR", "VAL", "UNK"};
       char Name[10];
       int i;

       if (resname[0] == '+') strcpy(Name, &resname[1]);
       else                   strcpy(Name, resname);

       for (i = 0; i < 30; i++)
            if (!strcmp(Name, standard_residue_list[i])) break;

       if (i < 30) return 1;
       else        return 0;
}

int is_aa_residue(char *resname)
{
       static char standard_aa_residue_list[170] = "ALA ARG ASN ASP CYS GLN GLU GLY HIS ILE LEU LYS MET PHE PRO SER THR TRP TYR VAL DAL DAR DAS DSP DCY DGN DGL DIL DLE DLY DPN DPR DSN DTH DTR DTY DVA";
       char *p = strstr(standard_aa_residue_list, resname);
       if (p) return 1;
       else   return 0;
}

void ndb_convert_time_format(char *indatestring, char *outdatestring, int fieldLength)
{
  char *buf;
  char mon[84], day[84], year[84], datestring[84];
  struct partime p;
  int i, j;
  char mons[12][84] =
    { "JAN", "FEB", "MAR", "APR", "MAY", "JUN",
        "JUL", "AUG", "SEP", "OCT", "NOV", "DEC"};

  if (!indatestring) {
    strcpy(outdatestring, "");
    return;
  }

  int exist_day = 1;
  if (strlen(indatestring) == 4 && isdigit(indatestring[0]) &&
      isdigit(indatestring[1]) && isdigit(indatestring[2]) && isdigit(indatestring[3])) {
    strcpy(outdatestring, indatestring);
    return;
  } else if (strlen(indatestring) == 8 && isalpha(indatestring[0]) && isalpha(indatestring[1]) &&
     isalpha(indatestring[2]) && indatestring[3] == '-' && isdigit(indatestring[4]) &&
     isdigit(indatestring[5]) && isdigit(indatestring[6]) && isdigit(indatestring[7])) {
    sprintf(datestring,"01-%s", indatestring);
    exist_day = 0;
  } else if (strlen(indatestring) == 6 && isalpha(indatestring[0]) && isalpha(indatestring[1]) &&
     isalpha(indatestring[2]) && indatestring[3] == '-' && isdigit(indatestring[4]) &&
     isdigit(indatestring[5])) {
    sprintf(datestring,"01-%s", indatestring);
    exist_day = 0;
  } else if (strlen(indatestring) == 7 && isdigit(indatestring[0]) && isdigit(indatestring[1]) &&
     isdigit(indatestring[2]) && isdigit(indatestring[3]) && indatestring[4] == '-' &&
     isdigit(indatestring[5]) && isdigit(indatestring[6])) {
    sprintf(datestring,"%s-01", indatestring);
    exist_day = 0;
  }  else {
    strcpy(datestring, indatestring);
  }

  for (i=0; i<(int) strlen(datestring); i++)
    if (!strncmp(&datestring[i], ":000",4)) break;
  if (i != (int) strlen(datestring)) {
    j = i+4;
    for (; i<j; i++)
      datestring[i] = ' ';
  }

  buf = partime(datestring, &p);
 
  if (strcmp(buf,"")) {
    ndb_log_message(NDB_MSG_ERR,"Date %s format error at  *%s*\n",
                    indatestring, buf);
  }
  if (p.tm.tm_mon < 0 || p.tm.tm_mon >= 12)
       strcpy(mon, mons[0]);
  else strcpy(mon, mons[p.tm.tm_mon]);

  if (p.tm.tm_mday > 0 && p.tm.tm_mday <= 31) {
    if (p.tm.tm_mday < 10)
         sprintf(day, "0%d", p.tm.tm_mday);
    else sprintf(day, "%d", p.tm.tm_mday);
  } else strcpy(day,  "01");

  if (fieldLength <= 10)  {
    if (p.tm.tm_year <0)
         sprintf(year, "00");
    else if (p.tm.tm_year % 100 < 10)
         sprintf(year, "0%d", p.tm.tm_year %100);
    else sprintf(year, "%d", p.tm.tm_year %100);
  } else {
    if (p.tm.tm_year > 1000)
         sprintf(year, "%d", p.tm.tm_year);
    else if (p.tm.tm_year > 100)
         sprintf(year, "1%d", p.tm.tm_year);
    else if (p.tm.tm_year > 10)
         sprintf(year, "19%d", p.tm.tm_year);
    else sprintf(year, "199%d", p.tm.tm_year);
  }
  if (exist_day)
       sprintf(outdatestring, "%s-%s-%s", day, mon, year);
  else sprintf(outdatestring, "%s-%s", mon, year);
}

int ndb_get_atom_number(char *atomName)
{
       char tmpAtom[84];

       ndb_get_atom_symbol(atomName, tmpAtom);
       if (!strcmp(tmpAtom, "D")) strcpy(tmpAtom, "H");
       int i = 0;
       for (i = 0; i < NUM_PERIODIC_TABLE_ATOM; i++) {
            if (!strcasecmp(tmpAtom, periodic_table[i].Atom_Symbol))
                break;
       }
       if (i == NUM_PERIODIC_TABLE_ATOM) return -1;
       else return (i + 1);
}

void ndb_get_atom_symbol(char *atomName, char *atomLabel)
{
       char tmpAtom[84], tmpAtom2[84];

       if (!isalnum(atomName[0]))
            strcpy(tmpAtom, &atomName[1]);
       else strcpy(tmpAtom, atomName);
       ndb_clean_string(tmpAtom);
       int str_len = strlen(tmpAtom);
       ndb_set_string_to_upper(tmpAtom);
       if (isdigit(tmpAtom[0]) && tmpAtom[1] == 'H') {
            strcpy(atomLabel, "H");
       }
       int i = 0;
       for (i = 0; i < str_len; i++) {
            if (!isdigit(tmpAtom[i]) && tmpAtom[i] != 'X' &&
                 tmpAtom[i] != ' ') break;
       }
       memset(tmpAtom2, 0, 3);
       strncpy(tmpAtom2, &tmpAtom[i],2);
       if (isdigit(tmpAtom2[1])) tmpAtom2[1] = 0;
       else {
            for (i = 0; i < NUM_PERIODIC_TABLE_ATOM; i++) {
            if (!strcasecmp(tmpAtom2, periodic_table[i].Atom_Symbol))
                 break;
            }
            if (i == NUM_PERIODIC_TABLE_ATOM) tmpAtom2[1] = 0;
       }
       strcpy(atomLabel, tmpAtom2);
}

void ndb_name_conversion_first_last(char *string)
{
       if (string == NULL || string[0] == '\0') return;

       char first[50], last[50];
       int i = 0;
       for (i = strlen(string) - 1; i >= 0; i--)
            if (string[i] == ',') break;

       if (i > 0) {
            strcpy(first, &string[i+1]);
            string[i] = '\0';
            strcpy(last, string);
            ndb_clean_string(first);
            ndb_clean_string(last);
            if (first[strlen(first)-1] == '.')
                 sprintf(string, "%s%s", first, last);
            else sprintf(string, "%s %s", first, last);
       }
}

void print_atom_atomName_field(const char *atomtype, const int aux_id, const int pos,
      char *temp_line, const char *atomName1, const char *residueName, int align)
{
       if (residueName && !strcmp(residueName, "HOH") && atomName1[0] == 'O') {
            sprintf(&temp_line[pos]," %c  ", 'O');
            return;
       }
       int l = strlen(atomName1);
       if (l == 4 || l == 0) {
            sprintf(&temp_line[pos], "%-4s", atomName1);
            return;
       }

       int i = 0, str_len = 0, n = 0, len = 0;
       char atomName[10], atom_type[5], *p = NULL;
       strcpy(atomName, atomName1);
       if (!strcmp(atomtype, "")) {
            ndb_clean_string(atomName);
            str_len = strlen(atomName);
            for (i = 0; i < str_len; i++)
                 if (!isdigit(atomName[i]) && atomName[i] != 'X') break;
            strncpy(atom_type, &atomName[i], 2);
            atom_type[2] = '\0';
            if (isdigit(atom_type[1])) atom_type[1] = '\0';
            if (aux_id == ATOMP_TOKEN-1) atom_type[1] = '\0';
            p = strstr(atomName, atom_type);
            n = p - atomName;
            len = strlen(atom_type);
            if (n == 1 || len == 2 || len == 1 && atomName[1] == atom_type[0] &&
                atom_type[0] != 'H' && atom_type[0] != 'C')
                 align = 0;
            else align = 1;
       }
       if (align == 0) {
            if (l == 2) sprintf(&temp_line[pos], "%2s  ", atomName);
            else if (l == 3) sprintf(&temp_line[pos], "%3s ", atomName);
       } else {
            if (l == 1) sprintf(&temp_line[pos], " %s  ", atomName);
            else if (l == 2) sprintf(&temp_line[pos], " %2s ", atomName);
            else if (l == 3) sprintf(&temp_line[pos], " %3s", atomName);
       }
}

void print_atom_card(FILE *fp, const int card_id, const int temp_id,
            int &atomSerialNo, int &WaterSerialNo, char Field[15][10],
            const int align)
{
       char temp_line[MxLen], FieldFormat[MxStrLen];
       int atomNameField = 2, residueNameField = 4;

       if (!(is_atom_card(temp_id) & AUX_ATOM)) {
            int atom_id = get_atom_card_id(temp_id);
            if (!strcmp(AtomInfo[atom_id].PdbTitle, "ATOM")) {
                 if (is_a_standard_residue(Field[4]))
                      sprintf(temp_line, "%-6s", AtomInfo[atom_id].PdbTitle);
                 else sprintf(temp_line, "%-6s", "HETATM");
            } else sprintf(temp_line, "%-6s", AtomInfo[atom_id].PdbTitle);
       } else sprintf(temp_line, "%-6s", NdbToken[temp_id].TokenName);

       if (!(is_atom_card(temp_id) & AUX_ATOM)) atomSerialNo++;

       int pos = 6, field_no = 0;
       for (field_no = 1; field_no < NdbToken[temp_id].NumField; field_no++) {
            if (NdbToken[temp_id].FieldList[field_no].NdbOnly) break;

            if ((field_no == 11 || field_no == 12) &&
                !strcmp("", Field[field_no])) {
                  if (field_no == 11) strcpy(Field[field_no], "1.0");
                  else strcpy(Field[field_no], "10.0");
            }
            ndb_set_string_to_upper(Field[field_no]);
            print_space_between_field(temp_id, field_no, temp_line, pos, 0);
            get_field_format(temp_id, field_no, FieldFormat, 0);
            if (field_no == 1) {
                 sprintf(Field[1], "%d", atomSerialNo);
                 sprintf(&temp_line[pos], FieldFormat, (double) atomSerialNo);
            } else if (strcmp("", Field[field_no])) {
                 if (field_no == atomNameField)
                      print_atom_atomName_field(Field[14], temp_id, pos, temp_line,
                          Field[field_no], Field[residueNameField], align);
                 else if (NdbToken[temp_id].FieldList[field_no].FieldType <= 2) {
                      double temp_double = atof(Field[field_no]);
                      sprintf(&temp_line[pos], FieldFormat, temp_double);
                 } else sprintf(&temp_line[pos], FieldFormat, Field[field_no]);
            } else {
                 sprintf(FieldFormat, "%%%-ds",
                          NdbToken[temp_id].FieldList[field_no].FieldWidth);
                 sprintf(&temp_line[pos], FieldFormat, "");
            }
            pos += NdbToken[temp_id].FieldList[field_no].FieldWidth;
       }

       sprintf(FieldFormat, "%%%ds", 
            78 - NdbToken[temp_id].FieldList[field_no].FieldStcol + 1);
       sprintf(&temp_line[pos], FieldFormat, "");
       pos = strlen(temp_line) - 1;
       int align_pos = 1;
       if (strlen(Field[14]) == 2) align_pos = 0;
       print_atom_atomName_field(Field[14], temp_id, pos, temp_line, Field[14],
               Field[residueNameField], align_pos);

       fprintf(fp, "%s\n", temp_line);
}

void print_jrnl_continuedf_field(const int jrnl_id, One_Card_Info_Format *pJrnlCard,
            const int pos, char *temp_line, char *FieldFormat)
{
       int i;
       char temp[MxStrLen], buffer[2000];
       
       int continueField = JrnlToken[jrnl_id].ContinuedF-1;
       pJrnlCard->CleanStringField(continueField);
       int str_len = strlen(pJrnlCard->Field(continueField));
       i = str_len;
       if (str_len > JrnlToken[jrnl_id].FieldList[continueField].FieldWidth) {
            if (strcmp(JrnlToken[jrnl_id].TokenName, "AUTH") !=0) {
                 for (i = JrnlToken[jrnl_id].FieldList[continueField].FieldWidth-1; i>=0; i--) {
                      if (pJrnlCard->Field(JrnlToken[jrnl_id].ContinuedF-1)[i] == ' ') break;
                 }
            }
            if (i < JrnlToken[jrnl_id].FieldList[continueField].FieldWidth / 5 ||
                strcmp(JrnlToken[jrnl_id].TokenName, "AUTH") == 0){
                 for (i = JrnlToken[jrnl_id].FieldList[continueField].FieldWidth-1; i>=0; i--) {
                      if (pJrnlCard->Field(continueField)[i] == ',' ||
                          pJrnlCard->Field(continueField)[i] == ')') break;
                      if (strcmp(JrnlToken[jrnl_id].TokenName,"AUTH") != 0 &&
                          pJrnlCard->Field(continueField)[i] == '-') break;
                 }
            }
            if (i < JrnlToken[jrnl_id].FieldList[continueField].FieldWidth / 5) 
                 i = JrnlToken[jrnl_id].FieldList[continueField].FieldWidth-1;
            else i = i + 1;
       } else i = str_len;
       strncpy(temp, pJrnlCard->Field(continueField), i);
       temp[i] = '\0';
       if (pJrnlCard->Field(continueField)[i] == ' ') i++;
       strcpy(buffer, &(pJrnlCard->Field(continueField)[i]));
       str_len = str_len - i ;
       if (str_len >= 0) buffer[str_len] = '\0';
       else buffer[0] = '\0';
       pJrnlCard->AddElementToField(buffer, continueField);
       sprintf(&temp_line[pos],FieldFormat, temp);
}

void print_ndb_continuedf_field(const int card_id, One_Card_Info_Format *pCard,
         const int pos, char *temp_line, char *FieldFormat, const int ContNo,
         const int PdbContNo)
{
       int i, str_len, max_len;
       char temp[200], buffer[4000];
       
       if (PdbContNo == 1) 
            max_len = NdbToken[card_id].FieldList[NdbToken[card_id].ContinuedF-1].FieldWidth;
       else max_len = NdbToken[card_id].FieldList[NdbToken[card_id].ContinuedF-1].FieldWidth-1;
     
       if (card_id == PCOMPN_TOKEN -1 || card_id == PSOURC_TOKEN-1)
            pCard->CleanStringField(NdbToken[card_id].ContinuedF-1);
       str_len = strlen(pCard->Field(NdbToken[card_id].ContinuedF-1));
     
       if (str_len > max_len) {
            for (i = max_len-1; i>=0; i--) {
                 if (pCard->Field(NdbToken[card_id].ContinuedF-1)[i] == ' ' ||
                     pCard->Field(NdbToken[card_id].ContinuedF-1)[i] == ',' ||
                     pCard->Field(NdbToken[card_id].ContinuedF-1)[i] == '-' ||
                     pCard->Field(NdbToken[card_id].ContinuedF-1)[i] == ')') 
                      break;
            }
            if (i < max_len / 5) i =  max_len;
            else i = i + 1;
       } else i = str_len;
     
       if (PdbContNo == 1 && NdbToken[card_id].PdbConField != 0 ||
           card_id == PCOMPN_TOKEN -1 || card_id == PSOURC_TOKEN-1 || 
           NdbToken[card_id].PdbConField == 0 && ContNo == 1) {
            strncpy(temp, pCard->Field(NdbToken[card_id].ContinuedF-1), i);
            temp[i] = '\0';
       } else {
            temp[0] = ' ';
            strncpy(&temp[1], pCard->Field(NdbToken[card_id].ContinuedF-1), i);
            temp[i+1] = '\0';
       }
       if (pCard->Field(NdbToken[card_id].ContinuedF-1)[i] == ' ') i++;
       strcpy(buffer, &(pCard->Field(NdbToken[card_id].ContinuedF-1)[i]));
       str_len = str_len - i ;
       if (str_len > 0) buffer[str_len] = '\0';
       else buffer[0] = '\0';
       pCard->AddElementToField(buffer, NdbToken[card_id].ContinuedF-1);
       sprintf(&temp_line[pos], FieldFormat, temp);
       for (i = 0; i < (int) strlen(temp_line); i++)
            if (temp_line[i] == '\n') temp_line[i] = ' ';
}

void print_space_between_field(const int card_id, const int field_no,
        char *temp_line, int &LinePos, const int IsJRNL)
{
       char temp_string[MxStrLen];
       if (IsJRNL == 0) {
            if (NdbToken[card_id].FieldList[field_no].FieldStcol != LinePos+1) {
                 sprintf(temp_string, "%%-%ds",
                         NdbToken[card_id].FieldList[field_no].FieldStcol -
                         NdbToken[card_id].FieldList[field_no-1].FieldStcol -
                         NdbToken[card_id].FieldList[field_no-1].FieldWidth);
                 sprintf(&temp_line[LinePos], temp_string, "");
                 LinePos = NdbToken[card_id].FieldList[field_no].FieldStcol -1;
            }
       } else {
            if (JrnlToken[card_id].FieldList[field_no].FieldStcol != LinePos+1){
                 sprintf(temp_string, "%%-%ds",
                       JrnlToken[card_id].FieldList[field_no].FieldStcol -
                       JrnlToken[card_id].FieldList[field_no-1].FieldStcol -
                       JrnlToken[card_id].FieldList[field_no-1].FieldWidth);
                 sprintf(&temp_line[LinePos], temp_string, "");
                 LinePos = JrnlToken[card_id].FieldList[field_no].FieldStcol-1;
            }
       }
}

void print_ter_card(FILE *fp, int &atomSerialNo, char Field[15][10])
{
       char FieldFormat[MxStrLen], temp_line[MxLen];
       int  residueNoField = 4;
       int  temp_id = find_card_id("TER");

       sprintf(temp_line, "%-6s","TER");
       int pos = 6;
       for (int field_no = 1; field_no < NdbToken[temp_id].NumField; field_no++) {
            ndb_set_string_to_upper(Field[field_no]);
            print_space_between_field(temp_id, field_no, temp_line, pos, 0);
            get_field_format(temp_id, field_no, FieldFormat, 0);
            if (field_no == residueNoField)
                 sprintf(FieldFormat, "%%%ds",
                      NdbToken[temp_id].FieldList[field_no].FieldWidth);
        
            if (field_no == 1) {
                 atomSerialNo++;
                 sprintf(&temp_line[pos], FieldFormat, (double) atomSerialNo);
            } else if (strcmp(Field[field_no+2], "") == 0){
                 sprintf(FieldFormat, "%%%ds",
                      NdbToken[temp_id].FieldList[field_no].FieldWidth);
                 sprintf(&temp_line[pos], FieldFormat, "");
            } else if (NdbToken[temp_id].FieldList[field_no].FieldType > 2 || 
               field_no == residueNoField)
                 sprintf(&temp_line[pos], FieldFormat, Field[field_no+2]);
            else {
                 double temp_double = atof(Field[field_no+2]);
                 sprintf(&temp_line[pos], FieldFormat, temp_double);
            }
            pos += NdbToken[temp_id].FieldList[field_no].FieldWidth;
       }
       fprintf(fp,"%s                                                     \n", temp_line);
}

